"""Base Class Functional."""

from __future__ import annotations

import math
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, TypeAlias

import torch
from typing_extensions import TypeVarTuple, Unpack

import mrpro.operators
from mrpro.operators.Operator import Operator

if TYPE_CHECKING:
    T = TypeVarTuple('T')
    FunctionalType: TypeAlias = Operator[Unpack[T], tuple[torch.Tensor]]
else:  # python 3.10 runtime compatibility. typing_extension
    FunctionalType: TypeAlias = Operator


def throw_if_negative_or_complex(
    x: torch.Tensor | complex, message: str = 'sigma must be real and contain only positive values'
) -> None:
    """Throw an ValueError if any element of x is negative or complex.

    Parameters
    ----------
    x
        input to be checked
    message
        error message that is raised if x contains negative or complex values
    """
    if (isinstance(x, float | int) and x >= 0) or (
        isinstance(x, torch.Tensor) and not x.dtype.is_complex and (x >= 0).all()
    ):
        return
    raise ValueError(message)


class ElementaryFunctional(Operator[torch.Tensor, tuple[torch.Tensor]], ABC):
    r"""Elementary functional base class.

    Here, an 'elementary' functional is a functional that can be written as
    :math:`f(x) = \phi ( \mathrm{weight} ( x - \mathrm{target}))`, returning a real value.
    It does not require another functional for initialization.
    """

    def __init__(
        self,
        target: torch.Tensor | None | complex = None,
        weight: torch.Tensor | complex = 1.0,
        dim: int | Sequence[int] | None = None,
        divide_by_n: bool = False,
        keepdim: bool = False,
    ) -> None:
        r"""Initialize a Functional.

        We assume that functionals are given in the form
        :math:`f(x) = \phi ( \mathrm{weight} ( x - \mathrm{target}))`
        for some functional :math:`\phi`.

        Parameters
        ----------
        target
            target element - often data tensor (see above)
        weight
            weight parameter (see above)
        dim
            dimension(s) over which functional is reduced.
            All other dimensions of  `weight ( x - target)` will be treated as batch dimensions.
        divide_by_n
            if true, the result is scaled by the number of elements of the dimensions index by `dim` in
            the tensor `weight ( x - target)`. If true, the functional is thus calculated as the mean,
            else the sum.
        keepdim
            if true, the dimension(s) of the input indexed by `dim` are maintained and collapsed to singletons,
            else they are removed from the result.

        """
        super().__init__()
        self.weight = torch.as_tensor(weight)
        if target is None:
            target = torch.tensor(0, dtype=torch.float32)
        self.target = torch.as_tensor(target)
        if isinstance(dim, int):
            dim = (dim,)
        elif isinstance(dim, Sequence):
            dim = tuple(dim)
        self.dim = dim
        self.divide_by_n = divide_by_n
        self.keepdim = keepdim

    def _divide_by_n(self, x: torch.Tensor, shape: None | Sequence[int]) -> torch.Tensor:
        """Apply factor for normalization.

        Input is scaled by the number of elements of either the input
        or product of the optional shape entries

        Parameters
        ----------
        x
            input to be scaled.
        shape
            input will be divided by the product of these numbers.
            If None, it will be divided by the number of elements of the input.

        Returns
        -------
            new scaled down tensor.
        """
        if not self.divide_by_n:
            return x
        if shape is None:
            shape = x.shape
        if self.dim is not None:
            size = [shape[i] for i in self.dim]
        else:
            size = list(shape)
        return x / math.prod(size)


class ProximableFunctional(Operator[torch.Tensor, tuple[torch.Tensor]], ABC):
    r"""ProximableFunctional Base Class.

    A proximable functional is a functional :math:`f(x)` that has a prox implementation,
    i.e. a function that yields :math:`\mathrm{argmin}_x \sigma f(x) + 1/2 ||x - y||_2^2`
    and a prox_convex_conjugate, yielding the prox of the convex conjugate.
    """

    @abstractmethod
    def prox(self, x: torch.Tensor, sigma: torch.Tensor | float = 1.0) -> tuple[torch.Tensor]:
        r"""Apply proximal operator.

        Yields :math:`\mathrm{prox}_{\sigma f}(x) = \mathrm{argmin}_{p} \sigma f(p) + 1/2 \|x-p\|_2^2` given :math:`x`
        and :math:`\sigma`.

        Parameters
        ----------
        x
            input tensor
        sigma
            scaling factor, must be positive

        Returns
        -------
            Proximal operator applied to the input tensor
        """

    def prox_convex_conj(self, x: torch.Tensor, sigma: torch.Tensor | float = 1.0) -> tuple[torch.Tensor]:
        r"""Apply proximal operator of convex conjugate of functional.

        Yields :math:`\mathrm{prox}_{\sigma f^*}(x) = \mathrm{argmin}_{p} \sigma f^*(p) + 1/2 \|x-p\|_2^2`,
        where :math:`f^*` denotes the convex conjugate of :math:`f`, given :math:`x` and :math:`\sigma`.

        Parameters
        ----------
        x
            input tensor
        sigma
            scaling factor, must be positive

        Returns
        -------
            Proximal operator  of the convex conjugate applied to the input tensor
        """
        if not isinstance(sigma, torch.Tensor):
            sigma = torch.as_tensor(1.0 * sigma)
        throw_if_negative_or_complex(sigma)
        sigma = sigma.clamp(min=1e-8)
        return (x - sigma * self.prox(x / sigma, 1 / sigma)[0],)

    def __rmul__(self, scalar: torch.Tensor | complex) -> ProximableFunctional:
        """Multiply functional with scalar."""
        if not isinstance(scalar, int | float | torch.Tensor):
            return NotImplemented
        return ScaledProximableFunctional(self, scalar)

    def __or__(
        self, other: ProximableFunctional
    ) -> mrpro.operators.ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor]:
        """Create a ProximableFunctionalSeparableSum from two proximable functionals.

        ``f | g`` is a separable sum with ``(f|g)(x,y) == f(x) + g(y)``.

        Parameters
        ----------
        other
            second functional to be summed

        Returns
        -------
            ProximableFunctionalSeparableSum object
        """
        if isinstance(other, ProximableFunctional):
            return mrpro.operators.ProximableFunctionalSeparableSum(self, other)
        return NotImplemented


class ElementaryProximableFunctional(ElementaryFunctional, ProximableFunctional):
    r"""Elementary proximable functional base class.

    Here, an 'elementary' functional is a functional that can be written as
    :math:`f(x) = \phi ( \mathrm{weight} ( x - \mathrm{target}))`, returning a real value.
    It does not require another functional for initialization.

    A proximable functional is a functional :math:`f(x)` that has a prox implementation,
    i.e. a function that yields :math:`\mathrm{argmin}_x \sigma f(x) + 1/2 \|x - y\|^2`.
    """


class ScaledProximableFunctional(ProximableFunctional):
    """Proximable Functional scaled by a scalar."""

    def __init__(self, functional: ProximableFunctional, scale: torch.Tensor | float) -> None:
        r"""Initialize a scaled proximable functional.

        A scaled functional is a functional that is scaled by a scalar factor :math:`\alpha`,
        i.e. :math:`f(x) = \alpha g(x)`.

        Parameters
        ----------
        functional
            proximable functional to be scaled
        scale
            scaling factor, must be real and positive
        """
        super().__init__()
        self.functional = functional
        self.scale = torch.as_tensor(scale)

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply the functional.

        Parameters
        ----------
        x
            input tensor

        Returns
        -------
            scaled output of the functional
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply the functional.

        .. note::
            Prefer calling the instance of the ScaledProximableFunctional as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return (self.scale * self.functional(x)[0],)

    def prox(self, x: torch.Tensor, sigma: torch.Tensor | float = 1.0) -> tuple[torch.Tensor]:
        """Proximal Mapping.

        Parameters
        ----------
        x
            input tensor
        sigma
            scaling factor

        Returns
        -------
            Proximal mapping applied to the input tensor
        """
        throw_if_negative_or_complex(
            self.scale, 'For prox to be defined, the scaling factor must be real and non-negative'
        )
        return (self.functional.prox(x, sigma * self.scale)[0],)

    def prox_convex_conj(self, x: torch.Tensor, sigma: torch.Tensor | float = 1.0) -> tuple[torch.Tensor]:
        """Proximal Mapping of the convex conjugate.

        Parameters
        ----------
        x
            input tensor
        sigma
            scaling factor

        Returns
        -------
            Proximal mapping of the convex conjugate applied to the input tensor
        """
        throw_if_negative_or_complex(
            self.scale, 'For prox_convex_conj to be defined, the scaling factor must be real and non-negative'
        )
        return (self.scale * self.functional.prox_convex_conj(x / self.scale, sigma / self.scale)[0],)



"""Differentiable Minimization."""

import functools
from collections.abc import Callable
from typing import TYPE_CHECKING, cast

import torch
from typing_extensions import Any, TypeVar, TypeVarTuple, Unpack

from mrpro.algorithms.optimizers.cg import cg
from mrpro.algorithms.optimizers.lbfgs import lbfgs
from mrpro.operators.Operator import Operator

ArgumentType = TypeVarTuple('ArgumentType')
VariableType = TypeVar('VariableType', bound=tuple[torch.Tensor, ...])
ObjectiveType = Callable[..., tuple[torch.Tensor]] | Operator[Any, tuple[torch.Tensor]]
FactoryType = Callable[..., ObjectiveType]
OptimizeFunctionType = Callable[[Callable, VariableType], VariableType]

default_lbfgs = functools.partial(
    lbfgs,
    learning_rate=1.0,
    max_iterations=40,
    tolerance_change=1e-6,
    tolerance_grad=1e-6,
    history_size=20,
    line_search_fn='strong_wolfe',
)
"""LBFGS Optimizer"""


class OptimizerCtx(torch.autograd.function.FunctionCtx):
    """Type hinting the CTX object."""

    factory: Callable[
        [Unpack[tuple[torch.Tensor, ...]]], Callable[[Unpack[tuple[torch.Tensor, ...]]], tuple[torch.Tensor]]
    ]
    len_x: int
    needs_input_grad: tuple[bool, ...]
    saved_tensors: tuple[torch.Tensor, ...]


class OptimizerFunction(torch.autograd.Function):
    """Implicit Backward."""

    if TYPE_CHECKING:

        @classmethod
        def apply(
            cls,
            factory: Callable[..., Callable[..., tuple[torch.Tensor]]],
            initial_values: tuple[torch.Tensor, ...],
            optimize: Callable[..., tuple[torch.Tensor, ...]] = default_lbfgs,
            *parameters: torch.Tensor,
        ) -> tuple[torch.Tensor, ...]:
            """Apply the function. Only used for type hinting."""
            return super().apply(factory, initial_values, optimize, *parameters)

    @staticmethod
    def forward(
        ctx: OptimizerCtx,
        factory: Callable[
            [Unpack[tuple[torch.Tensor, ...]]], Callable[[Unpack[tuple[torch.Tensor, ...]]], tuple[torch.Tensor]]
        ],
        initial_values: tuple[torch.Tensor, ...],
        optimize: Callable[..., tuple[torch.Tensor, ...]] = default_lbfgs,
        *parameters: torch.Tensor,
    ) -> tuple[torch.Tensor, ...]:
        """Optimize."""
        ctx.factory = factory

        parameters_ = tuple(p.detach().clone() if isinstance(p, torch.Tensor) else p for p in parameters)
        initial_values_ = tuple(
            x.detach().requires_grad_(True) if isinstance(x, torch.Tensor) else x for x in initial_values
        )
        objective = factory(*parameters_)
        solution = optimize(objective, initial_values)
        ctx.save_for_backward(*solution, *parameters_)
        ctx.len_x = len(initial_values_)
        return solution

    @staticmethod
    def backward(ctx: OptimizerCtx, *grad_outputs: torch.Tensor) -> tuple[torch.Tensor | None, ...]:
        """Calculate the backward pass using implicit differentiation."""
        solution = tuple(x.detach().clone().requires_grad_(True) for x in ctx.saved_tensors[: ctx.len_x])
        parameters = ctx.saved_tensors[ctx.len_x :]
        parameters = tuple(
            p.detach().clone().requires_grad_(True) if ctx.needs_input_grad[i + 3] else p.detach()
            for i, p in enumerate(parameters)
        )
        dparams = [p for p in parameters if p.requires_grad]

        objective = ctx.factory(*parameters)

        def hvp(*v: torch.Tensor) -> tuple[torch.Tensor, ...]:
            return torch.autograd.functional.vhp(lambda *x: objective(*x)[0], solution, v=v)[1]

        hessian_inverse_grad = cg(hvp, grad_outputs, max_iterations=50, tolerance=1e-6)
        with torch.enable_grad():
            dobjective_dsolution = torch.autograd.grad(objective(*solution), solution, create_graph=True)
            # - d^2_objective / d_solution d_params Hessian^-1 * grad
            grad_params = list(torch.autograd.grad(dobjective_dsolution, dparams, hessian_inverse_grad))
        grad_inputs: list[torch.Tensor | None] = [None, None, None]  # factory, x0, optimize
        for need_grad in ctx.needs_input_grad[3:]:
            if need_grad:
                grad_inputs.append(-grad_params.pop(0))
            else:
                grad_inputs.append(None)

        return tuple(grad_inputs)


class OptimizerOp(Operator[Unpack[ArgumentType], VariableType]):
    """Differentiable Optimization Operator.

    Finds :math:`x^*=argmin_x f_p(x)`.
    The solution :math:`x^*`  will be differentiable with respect to some parameters :math:`p`
    of the functional :math:`f`.

    One of the building blocks of PINQI [ZIMM2024]_


    References
    ----------
    .. [ZIMM2024] Zimmermann, Felix F., et al. (2024) PINQI. An End-to-End Physics-Informed Approach to Learned
       Quantitative MRI Reconstruction. IEEE TCI. https://doi.org/10.1109/TCI.2024.3388869
    """

    def __init__(
        self,
        factory: FactoryType,
        initializer: Callable[[Unpack[ArgumentType]], VariableType],
        optimize: OptimizeFunctionType = default_lbfgs,
    ):
        r"""Initialize a differentiable argmin solver.

        Parameters
        ----------
        factory
            Function, that given the parameters of the problem returns an objective function.
            The objective function should be a callable that takes the variable(s) as input and returns a scalar.
        initializer
            Function that, given the parameters of the problem, returns a tuple of initial values for the variable(s)
        optimize
            Function used to perform the optimization, for example `lbfgs`.
            Use `functools.partial` to setup up all settings besides the objective function and the initial values.

        Example
        -------
            Solving :math:`\|q(x)-y\|^2 + \alpha*\|x-x_\mathrm{reg}\|^2` with
            :math:`y`, :math:`\alpha` and :math:`x_\mathrm{reg}` parameters. The solution :math:`x^*` should be
            differentiable with respect to these.

            Use::

                def factory(y, alpha, x_reg):
                    return L2squared(y)@q+alpha*L2squared(x_reg)
                def initializer(_y, _alpha, _xreg):
                    return (x_reg,)
        """
        super().__init__()
        self.factory = factory
        self.optimize = optimize
        self.initializer = initializer

    def __call__(self, *parameters: Unpack[ArgumentType]) -> VariableType:
        """Find the argmin.

        Parameters
        ----------
        parameters
            Parameters of the argmin problem. These will be passed to the factory functions defined in the initializer.

        Returns
        -------
            The solution of the argmin problem.
        """
        return super().__call__(*parameters)

    def forward(self, *parameters: Unpack[ArgumentType]) -> VariableType:
        """Find the argmin.

        .. note::
            Prefer calling the instance of the OptimizerOp as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        initial_values = self.initializer(*parameters)
        initial_values_ = tuple(x.clone() if any(x is p for p in parameters) else x for x in initial_values)
        result = OptimizerFunction.apply(
            self.factory, initial_values_, self.optimize, *cast(tuple[torch.Tensor, ...], parameters)
        )
        return cast(VariableType, result)



"""Endomorph Operators."""

from __future__ import annotations

from abc import abstractmethod
from typing import cast

import torch
from typing_extensions import Any, TypeVar, TypeVarTuple, Unpack, overload

import mrpro.operators
from mrpro.operators.Operator import Operator
from mrpro.utils.typing import endomorph

Tin = TypeVarTuple('Tin')
Tout = TypeVar('Tout', bound=tuple[torch.Tensor, ...], covariant=True)


class EndomorphOperator(Operator[Unpack[tuple[torch.Tensor, ...]], tuple[torch.Tensor, ...]]):
    """Endomorph Operator.

    Endomorph Operators have N tensor inputs and exactly N outputs.
    """

    @endomorph
    def __call__(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply the endomorphism.

        An endomorphism is an operator that maps a set of tensors
        to another set of tensors of the same number.

        Parameters
        ----------
        *x
            One or more input tensors.

        Returns
        -------
            A tuple containing the same number of tensors as input,
            resulting from the operator's action.
        """
        # This function only overwrites the type hints of the base operator class
        return super().__call__(*x)

    @abstractmethod
    @endomorph
    def forward(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply forward of EndomorphOperator.

        .. note::
            Prefer calling the instance of the EndomorphOperator operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """

    @overload
    def __matmul__(self, other: EndomorphOperator) -> EndomorphOperator: ...
    @overload
    def __matmul__(self, other: Operator[Unpack[Tin], Tout]) -> Operator[Unpack[Tin], Tout]: ...

    def __matmul__(
        self, other: Operator[Unpack[Tin], Tout] | EndomorphOperator
    ) -> Operator[Unpack[Tin], Tout] | EndomorphOperator:
        """Operator composition."""
        if isinstance(other, mrpro.operators.MultiIdentityOp):
            return self
        elif isinstance(self, mrpro.operators.MultiIdentityOp):
            return other

        res = super().__matmul__(cast(Any, other))  # avoid mypy 1.11 crash
        if isinstance(other, EndomorphOperator):
            return cast(EndomorphOperator, res)
        else:
            return cast(Operator[Unpack[Tin], Tout], res)

    def __rmatmul__(self, other: Operator[Unpack[Tin], Tout]) -> Operator[Unpack[Tin], Tout]:
        """Operator composition."""
        return other.__matmul__(cast(Operator[Unpack[Tin], tuple[Unpack[Tin]]], self))



"""Transient steady state signal models."""

from collections.abc import Sequence

import torch

from mrpro.operators.SignalModel import SignalModel
from mrpro.utils.reshape import unsqueeze_right


class TransientSteadyStateWithPreparation(SignalModel[torch.Tensor, torch.Tensor, torch.Tensor]):
    r"""Signal model for transient steady state.

    This signal model describes the behavior of the longitudinal magnetization during continuous acquisition after
    a preparation pulse. The effect of the preparation pulse is modeled by a scaling factor applied to the
    equilibrium magnetization. A delay after the preparation pulse can be defined. During this time T1 relaxation to M0
    occurs. Data acquisition starts after this delay. Perfect spoiling is assumed and hence T2 effects are not
    considered in the model. In addition this model assumes :math:`TR << T1` and :math:`TR << T1^*`
    (see definition below) [DEI1992]_ [LOO1970]_.

    Let's assume we want to describe a continuous acquisition after an inversion pulse, then we have three parts:
    [Part A: 180° inversion pulse][Part B: spoiler gradient][Part C: Continuous data acquisition]

    - Part A: The 180° pulse leads to an inversion of the equilibrium magnetization (:math:`M_0`) to :math:`-M_0`.
      This can be described by setting the scaling factor `m0_scaling_preparation` to `-1`.

    - Part B: Commonly after an inversion pulse a strong spoiler gradient is played out to compensate for non-perfect
      inversion. During this time the magnetization :math:`M_z(t)` follows the signal model:
      :math:`M_z(t) = M_0 + (s * M_0 - M_0)e^{(-t / T1)}` where :math:`s` is `m0_scaling_preparation`.

    - Part C: After the spoiler gradient the data acquisition starts and the magnetization :math:`M_z(t)` can be
      described by the signal model: :math:`M_z(t) = M_0^* + (M_{init} - M_0^*)e^{(-t / T1^*)}`
      where the initial magnetization is :math:`M_{init} = M_0 + (s*M_0 - M_0)e^{(-\Delta t / T1)}`,
      where :math:`s` is `m0_scaling_preparation` and :math:`\Delta t` is `delay_after_preparation`.
      The effective longitudinal relaxation time is :math:`T1^* = 1/(1/T1 - ln(cos(\alpha)/TR)`
      where :math:`TR` is `repetition_time` and :math:`\alpha` is `flip_angle`.
      The steady-state magnetization is :math:`M_0^* = M_0 T1^* / T1`.

    References
    ----------
    .. [DEI1992] Deichmann R, Haase A (1992) Quantification of T1 values by SNAPSHOT-FLASH NMR imaging. J. Magn. Reson.
       612 http://doi.org/10.1016/0022-2364(92)90347-A
    .. [LOO1970] Look D, Locker R (1970) Time Saving in Measurement of NMR and EPR Relaxation Times. Rev. Sci. Instrum
       41 https://doi.org/10.1063/1.1684482
    """

    def __init__(
        self,
        sampling_time: float | torch.Tensor | Sequence[float],
        repetition_time: float | torch.Tensor,
        m0_scaling_preparation: float | torch.Tensor = 1.0,
        delay_after_preparation: float | torch.Tensor = 0.0,
    ) -> None:
        """Initialize transient steady state signal model.

        `repetition_time`, `m0_scaling_preparation` and `delay_after_preparation` can vary for each voxel and will be
        broadcasted starting from the front (i.e. from the other dimension).

        Parameters
        ----------
        sampling_time
            Time points when model is evaluated. A `sampling_time` of 0 describes the first acquired data point
            after the inversion pulse and spoiler gradients. To take the T1 relaxation during the delay between
            inversion pulse and start of data acquisition into account, set the `delay_after_preparation` > 0.
            with shape `(time, ...)`
        repetition_time
            repetition time
        m0_scaling_preparation
            Scaling of the equilibrium magnetization due to the preparation pulse before the data acquisition.
        delay_after_preparation
            Time between preparation pulse and start of data acquisition. During this time, standard longitudinal
            relaxation occurs.

        """
        super().__init__()

        sampling_time_tensor = torch.as_tensor(sampling_time)
        self.sampling_time = torch.nn.Parameter(sampling_time_tensor, requires_grad=sampling_time_tensor.requires_grad)

        repetition_time_tensor = torch.as_tensor(repetition_time, device=sampling_time_tensor.device)
        self.repetition_time = torch.nn.Parameter(
            repetition_time_tensor, requires_grad=repetition_time_tensor.requires_grad
        )

        m0_scaling_preparation_tensor = torch.as_tensor(m0_scaling_preparation, device=sampling_time_tensor.device)
        self.m0_scaling_preparation = torch.nn.Parameter(
            m0_scaling_preparation_tensor, requires_grad=m0_scaling_preparation_tensor.requires_grad
        )

        delay_after_preparation_tensor = torch.as_tensor(delay_after_preparation, device=sampling_time_tensor.device)
        self.delay_after_preparation = torch.nn.Parameter(
            delay_after_preparation_tensor, requires_grad=delay_after_preparation_tensor.requires_grad
        )

    def __call__(self, m0: torch.Tensor, t1: torch.Tensor, flip_angle: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the transient steady-state signal model with preparation.

        Calculates the signal based on the formula:
        :math:`M_z(t) = M_0^* + (M_{init} - M_0^*) * exp(-t / T_1^*)`,
        where :math:`M_{init}` is the magnetization after preparation and initial delay,
        :math:`M_0^*` is the effective steady-state magnetization, and :math:`T_1^*` is the
        effective T1 relaxation time during continuous acquisition.

        Parameters
        ----------
        m0
            Equilibrium signal / proton density.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        t1
            Longitudinal relaxation time T1.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        flip_angle
            Flip angle of data acquisition rf pulses in radians.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.

        Returns
        -------
            Signal calculated for each sampling time.
            Shape `(times ...)`, for example `(times, *other, coils, z, y, x)`, or `(times, samples)`
            where `times` is the number of sampling times.
        """
        return super().__call__(m0, t1, flip_angle)

    def forward(self, m0: torch.Tensor, t1: torch.Tensor, flip_angle: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of TransientSteadyStateWithPreparation.

        .. note::
            Prefer calling the instance of the TransientSteadyStateWithPreparation as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        ndim = max(m0.ndim, t1.ndim, flip_angle.ndim)
        repetition_time = unsqueeze_right(self.repetition_time, ndim - self.repetition_time.ndim)
        m0_scaling_preparation = unsqueeze_right(self.m0_scaling_preparation, ndim - self.m0_scaling_preparation.ndim)
        delay_after_preparation = unsqueeze_right(
            self.delay_after_preparation, ndim - self.delay_after_preparation.ndim
        )
        # leftmost is time
        sampling_time = unsqueeze_right(self.sampling_time, m0.ndim - self.sampling_time.ndim + 1)

        # effect of preparation pulse
        m_start = m0 * m0_scaling_preparation

        # relaxation towards M0
        m_start = m0 + (m_start - m0) * torch.exp(-(delay_after_preparation / t1))

        # transient steady state
        ln_cos_tr = torch.log(torch.cos(flip_angle)) / repetition_time
        r1_star = 1 / t1 - ln_cos_tr
        m0_star = m0 / (1 - t1 * ln_cos_tr)
        signal = m0_star + (m_start - m0_star) * torch.exp(-sampling_time * r1_star)
        return (signal,)



"""Cartesian trajectory class."""

from collections.abc import Sequence

import torch

from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator
from mrpro.utils.RandomGenerator import RandomGenerator
from mrpro.utils.reshape import unsqueeze_tensors_left


class KTrajectoryCartesian(KTrajectoryCalculator):
    """Cartesian trajectory."""

    def __call__(
        self,
        *,
        n_k0: int,
        k0_center: int | torch.Tensor,
        k1_idx: torch.Tensor,
        k1_center: int | torch.Tensor,
        k2_idx: torch.Tensor,
        k2_center: int | torch.Tensor,
        reversed_readout_mask: torch.Tensor | None = None,
        **_,
    ) -> KTrajectory:
        """Calculate Cartesian trajectory for given KHeader.

        Parameters
        ----------
        n_k0
            number of samples in k0
        k0_center
            position of k-space center in k0
        k1_idx
            indices of k1
        k1_center
            position of k-space center in k1
        k2_idx
            indices of k2
        k2_center
            position of k-space center in k2
        reversed_readout_mask
            boolean tensor indicating reversed readout

        Returns
        -------
            Cartesian trajectory for given KHeader
        """
        # K-space locations along readout lines
        kx = self._readout(n_k0, k0_center, reversed_readout_mask=reversed_readout_mask)

        # Trajectory along phase and slice encoding
        ky = (k1_idx - k1_center).to(torch.float32)
        kz = (k2_idx - k2_center).to(torch.float32)

        kz, ky, kx = unsqueeze_tensors_left(kz, ky, kx, ndim=5)
        return KTrajectory(kz, ky, kx)

    @classmethod
    def fullysampled(cls, encoding_matrix: SpatialDimension[int]) -> KTrajectory:
        """Generate fully sampled Cartesian trajectory.

        Parameters
        ----------
        encoding_matrix
            Encoded K-space size.

        Returns
        -------
            Cartesian trajectory.
        """
        return cls()(
            n_k0=encoding_matrix.x,
            k0_center=encoding_matrix.x // 2,
            k1_idx=torch.arange(encoding_matrix.y)[:, None],
            k1_center=encoding_matrix.y // 2,
            k2_idx=torch.arange(encoding_matrix.z)[:, None, None],
            k2_center=encoding_matrix.z // 2,
        )

    @classmethod
    def gaussian_variable_density(
        cls,
        encoding_matrix: SpatialDimension[int] | int,
        acceleration: float = 2.0,
        n_center: int = 10,
        fwhm_ratio: float = 1.0,
        n_other: Sequence[int] = (1,),
        seed: int | None = None,
    ) -> KTrajectory:
        """
        Generate k-space Gaussian weighted variable density sampling.

        Parameters
        ----------
        encoding_matrix
            Encoded K-space size, must have ``encoding_matrix.z=1``.
            If a single integer, a square k-space is considered.
        acceleration
            Acceleration factor (undersampling rate).
        n_center
            Number of fully-sampled center lines to always include.
        fwhm_ratio
            Full-width at half-maximum of the Gaussian relative to encoding_matrix.y.
            Larger values approach uniform sampling. Set to infinity for uniform sampling.
        n_other
            Batch size(s). The trajectory is different for each batch sample.
        seed
            Random seed for reproducibility.


        Returns
        -------
            Cartesian trajectory.

        Raises
        ------
        ValueError
            If `n_center` exceeds the total number of lines to keep given the acceleration.
        NotImplementedError
            If called with a 3D encoding matrix.
        """
        if isinstance(encoding_matrix, int):
            n_k1 = encoding_matrix
            n_k0 = encoding_matrix
        elif encoding_matrix.z > 1:
            raise NotImplementedError('Only 2D random trajectories can be created this way.')
        else:
            n_k1, n_k0 = encoding_matrix.y, encoding_matrix.x

        n_keep = min(int(n_k1 / acceleration), n_k1)
        if n_center > n_keep:
            raise ValueError(f'Number of center lines ({n_center}) exceeds number of lines to keep ({n_keep}).')
        rng = RandomGenerator(seed)
        k1_idx = rng.gaussian_variable_density_samples(
            (*n_other, n_keep),
            low=-n_k1 // 2,
            high=n_k1 // 2,
            fwhm=fwhm_ratio * n_k1,
            always_sample=range(-n_center // 2, n_center // 2),
        )

        return cls()(
            n_k0=n_k0,
            k0_center=n_k0 // 2,
            k1_idx=k1_idx[..., None, None, :, None],
            k1_center=0,
            k2_idx=torch.tensor(0),
            k2_center=0,
        )



"""Algorithms for reconstructions, optimization, density and sensitivity map estimation, etc."""

from mrpro.algorithms import csm, optimizers, reconstruction, dcf
from mrpro.algorithms.prewhiten_kspace import prewhiten_kspace
from mrpro.algorithms.total_variation_denoising import total_variation_denoising
__all__ = ["csm", "dcf", "optimizers", "prewhiten_kspace", "reconstruction", "total_variation_denoising"]



"""Test ISMRMRD acquisitions based on their flags."""

import ismrmrd

from mrpro.data.enums import AcqFlags

# Same criteria as https://github.com/wtclarke/pymapvbvd/blob/master/mapvbvd/mapVBVD.py uses
DEFAULT_IGNORE_FLAGS = (
    AcqFlags.ACQ_IS_NOISE_MEASUREMENT
    | AcqFlags.ACQ_IS_DUMMYSCAN_DATA
    | AcqFlags.ACQ_IS_HPFEEDBACK_DATA
    | AcqFlags.ACQ_IS_NAVIGATION_DATA
    | AcqFlags.ACQ_IS_PHASECORR_DATA
    | AcqFlags.ACQ_IS_PHASE_STABILIZATION
    | AcqFlags.ACQ_IS_PHASE_STABILIZATION_REFERENCE
    | AcqFlags.ACQ_IS_PARALLEL_CALIBRATION
)


def is_image_acquisition(acquisition: ismrmrd.Acquisition) -> bool:
    """Test if acquisition was obtained for imaging.

    Parameters
    ----------
    acquisition
        ISMRMRD acquisition

    Returns
    -------
        True if the acquisition was obtained for imaging
    """
    return not DEFAULT_IGNORE_FLAGS.value & acquisition.flags


def is_noise_acquisition(acquisition: ismrmrd.Acquisition) -> bool:
    """Test if acquisition contains noise measurements.

    Parameters
    ----------
    acquisition
        ISMRMRD acquisition

    Returns
    -------
        True if the acquisition contains a noise measurement
    """
    return AcqFlags.ACQ_IS_NOISE_MEASUREMENT.value & acquisition.flags


def is_coil_calibration_acquisition(acquisition: ismrmrd.Acquisition) -> bool:
    """Test if acquisitions was obtained for coil calibration.

    Parameters
    ----------
    acquisition
        ISMRMRD acquisition

    Returns
    -------
        True if the acquisition contains coil calibration data
    """
    coil_calibration_flag = AcqFlags.ACQ_IS_PARALLEL_CALIBRATION | AcqFlags.ACQ_IS_PARALLEL_CALIBRATION_AND_IMAGING
    return coil_calibration_flag.value & acquisition.flags


def has_n_coils(n_coils: int, acquisition: ismrmrd.Acquisition) -> bool:
    """Test if acquisitions was obtained with a certain number of receiver coils.

    Parameters
    ----------
    n_coils
        number of receiver coils
    acquisition
        ISMRMRD acquisition

    Returns
    -------
        True if the acquisition was obtained with n_coils receiver coils
    """
    return acquisition.data.shape[0] == n_coils



"""Acquisition information dataclass."""

from collections.abc import Callable, Sequence
from dataclasses import field, fields
from typing import Literal, TypeAlias, overload

import ismrmrd
import numpy as np
import torch
from typing_extensions import Self

from mrpro.data.Dataclass import Dataclass
from mrpro.data.Rotation import Rotation
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.utils.reshape import unsqueeze_at, unsqueeze_right
from mrpro.utils.unit_conversion import m_to_mm, micrometer_to_m, mm_to_m

_convert_time_stamp_type: TypeAlias = Callable[
    [
        torch.Tensor,
        Literal[
            'acquisition_time_stamp', 'physiology_time_stamp_1', 'physiology_time_stamp_2', 'physiology_time_stamp_3'
        ],
    ],
    torch.Tensor,
]


def convert_time_stamp_from_siemens(
    timestamp: torch.Tensor,
    _: str,
) -> torch.Tensor:
    """Convert Siemens time stamp to seconds."""
    return timestamp.double() * 2.5e-3


def convert_time_stamp_to_siemens(
    timestamp: float,
) -> int:
    """Convert time stamp in seconds to Siemens time steps (2.5ms)."""
    return int(timestamp / 2.5e-3)


def convert_time_stamp_from_osi2(
    timestamp: torch.Tensor,
    _: str,
) -> torch.Tensor:
    """Convert OSI2 time stamp to seconds."""
    return timestamp.double() * 1e-3


def convert_time_stamp_to_osi2(
    timestamp: float,
) -> int:
    """Convert time stamp in seconds to OSI2 time steps (1ms)."""
    return int(timestamp * 1e3)


def _int_factory() -> torch.Tensor:
    return torch.zeros(1, 1, 1, 1, 1, dtype=torch.int64)


def _float_factory() -> torch.Tensor:
    return torch.zeros(1, 1, 1, 1, 1, dtype=torch.float)


def _position_factory() -> SpatialDimension[torch.Tensor]:
    return SpatialDimension(
        torch.zeros(1, 1, 1, 1, 1, dtype=torch.float),
        torch.zeros(1, 1, 1, 1, 1, dtype=torch.float),
        torch.zeros(1, 1, 1, 1, 1, dtype=torch.float),
    )


class AcqIdx(Dataclass):
    """Acquisition index for each readout."""

    k1: torch.Tensor = field(default_factory=_int_factory)
    """First phase encoding."""

    k2: torch.Tensor = field(default_factory=_int_factory)
    """Second phase encoding."""

    average: torch.Tensor = field(default_factory=_int_factory)
    """Signal average."""

    slice: torch.Tensor = field(default_factory=_int_factory)
    """Slice number (multi-slice 2D)."""

    contrast: torch.Tensor = field(default_factory=_int_factory)
    """Echo number in multi-echo."""

    phase: torch.Tensor = field(default_factory=_int_factory)
    """Cardiac phase."""

    repetition: torch.Tensor = field(default_factory=_int_factory)
    """Counter in repeated/dynamic acquisitions."""

    set: torch.Tensor = field(default_factory=_int_factory)
    """Sets of different preparation, e.g. flow encoding, diffusion weighting."""

    segment: torch.Tensor = field(default_factory=_int_factory)
    """Counter for segmented acquisitions."""

    user0: torch.Tensor = field(default_factory=_int_factory)
    """User index 0."""

    user1: torch.Tensor = field(default_factory=_int_factory)
    """User index 1."""

    user2: torch.Tensor = field(default_factory=_int_factory)
    """User index 2."""

    user3: torch.Tensor = field(default_factory=_int_factory)
    """User index 3."""

    user4: torch.Tensor = field(default_factory=_int_factory)
    """User index 4."""

    user5: torch.Tensor = field(default_factory=_int_factory)
    """User index 5."""

    user6: torch.Tensor = field(default_factory=_int_factory)
    """User index 6."""

    user7: torch.Tensor = field(default_factory=_int_factory)
    """User index 7."""

    def __post_init__(self) -> None:
        """Ensure that all indices are broadcastable."""
        f = [getattr(self, field.name) for field in fields(self)]
        try:
            torch.broadcast_shapes(*[field.shape for field in f])
        except RuntimeError:
            raise ValueError('The acquisition index dimensions must be broadcastable.') from None
        if any(x.ndim < 5 for x in f):
            raise ValueError('The acquisition index tensors should each have at least 5 dimensions.')


class UserValues(Dataclass):
    """User Values used in AcqInfo."""

    float0: torch.Tensor = field(default_factory=_float_factory)
    float1: torch.Tensor = field(default_factory=_float_factory)
    float2: torch.Tensor = field(default_factory=_float_factory)
    float3: torch.Tensor = field(default_factory=_float_factory)
    float4: torch.Tensor = field(default_factory=_float_factory)
    float5: torch.Tensor = field(default_factory=_float_factory)
    float6: torch.Tensor = field(default_factory=_float_factory)
    float7: torch.Tensor = field(default_factory=_float_factory)
    int0: torch.Tensor = field(default_factory=_int_factory)
    int1: torch.Tensor = field(default_factory=_int_factory)
    int2: torch.Tensor = field(default_factory=_int_factory)
    int3: torch.Tensor = field(default_factory=_int_factory)
    int4: torch.Tensor = field(default_factory=_int_factory)
    int5: torch.Tensor = field(default_factory=_int_factory)
    int6: torch.Tensor = field(default_factory=_int_factory)
    int7: torch.Tensor = field(default_factory=_int_factory)


class PhysiologyTimestamps(Dataclass):
    """Time stamps relative to physiological triggering, e.g. ECG, in seconds."""

    timestamp0: torch.Tensor = field(default_factory=_float_factory)
    timestamp1: torch.Tensor = field(default_factory=_float_factory)
    timestamp2: torch.Tensor = field(default_factory=_float_factory)


class AcqInfo(Dataclass):
    """Acquisition information for each readout."""

    idx: AcqIdx = field(default_factory=AcqIdx)
    """Indices describing acquisitions (i.e. readouts)."""

    acquisition_time_stamp: torch.Tensor = field(default_factory=_float_factory)
    """Clock time stamp [s] (Siemens: seconds since midnight)"""

    flags: torch.Tensor = field(default_factory=_int_factory)
    """A bit mask of common attributes applicable to individual acquisition readouts."""

    orientation: Rotation = field(default_factory=lambda: Rotation.identity((1, 1, 1, 1, 1)))
    """Rotation describing the orientation of the readout, phase and slice encoding direction."""

    patient_table_position: SpatialDimension[torch.Tensor] = field(default_factory=_position_factory)
    """Offset position of the patient table, in LPS coordinates [m]."""

    physiology_time_stamps: PhysiologyTimestamps = field(default_factory=PhysiologyTimestamps)
    """Time stamps relative to physiological triggering, e.g. ECG [s]."""

    position: SpatialDimension[torch.Tensor] = field(default_factory=_position_factory)
    """Center of the excited volume, in LPS coordinates relative to isocenter [m]."""

    sample_time_us: torch.Tensor = field(default_factory=_float_factory)
    """Readout bandwidth, as time between samples [us]."""

    user: UserValues = field(default_factory=UserValues)
    """User defined float or int values"""

    @overload
    @classmethod
    def from_ismrmrd_acquisitions(
        cls,
        acquisitions: Sequence[ismrmrd.acquisition.Acquisition],
        *,
        additional_fields: None,
        convert_time_stamp: _convert_time_stamp_type = convert_time_stamp_from_siemens,
    ) -> Self: ...

    @overload
    @classmethod
    def from_ismrmrd_acquisitions(
        cls,
        acquisitions: Sequence[ismrmrd.acquisition.Acquisition],
        *,
        additional_fields: Sequence[str],
        convert_time_stamp: _convert_time_stamp_type = convert_time_stamp_from_siemens,
    ) -> tuple[Self, tuple[torch.Tensor, ...]]: ...

    @classmethod
    def from_ismrmrd_acquisitions(
        cls,
        acquisitions: Sequence[ismrmrd.acquisition.Acquisition],
        *,
        additional_fields: Sequence[str] | None = None,
        convert_time_stamp: _convert_time_stamp_type = convert_time_stamp_from_siemens,
    ) -> Self | tuple[Self, tuple[torch.Tensor, ...]]:
        """Read the header of a list of acquisition and store information.

        Parameters
        ----------
        acquisitions
            list of ismrmrd acquisistions to read from. Needs at least one acquisition.
        additional_fields
            if supplied, additional information from fields with these names will be extracted from the
            ismrmrd acquisitions and returned as tensors.
        convert_time_stamp
            function used to convert the raw time stamps to seconds.
        """
        # Idea: create array of structs, then a struct of arrays,
        # convert it into tensors to store in our dataclass.

        if len(acquisitions) == 0:
            raise ValueError('Acquisition list must not be empty.')

        # Creating the dtype first and casting to bytes
        # is a workaround for a bug in cpython causing a warning
        # if np.array(AcquisitionHeader) is called directly.
        # also, this needs to check the dtype only once.
        acquisition_head_dtype = np.dtype(ismrmrd.AcquisitionHeader)
        headers = np.frombuffer(
            np.array([memoryview(a._head).cast('B') for a in acquisitions]),
            dtype=acquisition_head_dtype,
        )

        idx = headers['idx']

        def tensor(data: np.ndarray) -> torch.Tensor:
            # we have to convert first as pytoch cant create tensors from np.uint16 arrays
            # we use int32 for uint16 and int64 for uint32 to fit largest values.
            match data.dtype:
                case np.uint16:
                    data = data.astype(np.int32)  # type: ignore[unreachable]
                case np.uint32 | np.uint64:
                    data = data.astype(np.int64)  # type: ignore[unreachable]
            # Remove any unnecessary dimensions
            return torch.tensor(np.squeeze(data))

        def tensor_5d(data: np.ndarray) -> torch.Tensor:
            # Convert tensor to torch dtypes and ensure it is 5D
            data_tensor = tensor(data)
            return unsqueeze_right(data_tensor, 5 - data_tensor.ndim)

        def spatialdimension_5d(data: np.ndarray) -> SpatialDimension[torch.Tensor]:
            data_tensor = torch.tensor(data, dtype=torch.float32)
            data_tensor = unsqueeze_at(data_tensor, -2, 5 - data_tensor.ndim + 1)
            # all spatial dimensions are float32
            return SpatialDimension.from_array_xyz(data_tensor)

        acq_idx = AcqIdx(
            k1=tensor_5d(idx['kspace_encode_step_1']),
            k2=tensor_5d(idx['kspace_encode_step_2']),
            average=tensor_5d(idx['average']),
            slice=tensor_5d(idx['slice']),
            contrast=tensor_5d(idx['contrast']),
            phase=tensor_5d(idx['phase']),
            repetition=tensor_5d(idx['repetition']),
            set=tensor_5d(idx['set']),
            segment=tensor_5d(idx['segment']),
            user0=tensor_5d(idx['user'][:, 0]),
            user1=tensor_5d(idx['user'][:, 1]),
            user2=tensor_5d(idx['user'][:, 2]),
            user3=tensor_5d(idx['user'][:, 3]),
            user4=tensor_5d(idx['user'][:, 4]),
            user5=tensor_5d(idx['user'][:, 5]),
            user6=tensor_5d(idx['user'][:, 6]),
            user7=tensor_5d(idx['user'][:, 7]),
        )
        user = UserValues(
            tensor_5d(headers['user_float'][:, 0]),
            tensor_5d(headers['user_float'][:, 1]),
            tensor_5d(headers['user_float'][:, 2]),
            tensor_5d(headers['user_float'][:, 3]),
            tensor_5d(headers['user_float'][:, 4]),
            tensor_5d(headers['user_float'][:, 5]),
            tensor_5d(headers['user_float'][:, 6]),
            tensor_5d(headers['user_float'][:, 7]),
            tensor_5d(headers['user_int'][:, 0]),
            tensor_5d(headers['user_int'][:, 1]),
            tensor_5d(headers['user_int'][:, 2]),
            tensor_5d(headers['user_int'][:, 3]),
            tensor_5d(headers['user_int'][:, 4]),
            tensor_5d(headers['user_int'][:, 5]),
            tensor_5d(headers['user_int'][:, 6]),
            tensor_5d(headers['user_int'][:, 7]),
        )
        physiology_time_stamps = PhysiologyTimestamps(
            convert_time_stamp(tensor_5d(headers['physiology_time_stamp'][:, 0]), 'physiology_time_stamp_1'),
            convert_time_stamp(tensor_5d(headers['physiology_time_stamp'][:, 1]), 'physiology_time_stamp_2'),
            convert_time_stamp(tensor_5d(headers['physiology_time_stamp'][:, 2]), 'physiology_time_stamp_3'),
        )
        acq_info = cls(
            idx=acq_idx,
            acquisition_time_stamp=convert_time_stamp(
                tensor_5d(headers['acquisition_time_stamp']), 'acquisition_time_stamp'
            ),
            flags=tensor_5d(headers['flags']),
            orientation=Rotation.from_directions(
                spatialdimension_5d(headers['slice_dir']),
                spatialdimension_5d(headers['phase_dir']),
                spatialdimension_5d(headers['read_dir']),
            ),
            patient_table_position=spatialdimension_5d(headers['patient_table_position']).apply_(micrometer_to_m),
            position=spatialdimension_5d(headers['position']).apply_(mm_to_m),
            sample_time_us=tensor_5d(headers['sample_time_us']),
            user=user,
            physiology_time_stamps=physiology_time_stamps,
        )

        if additional_fields is None:
            return acq_info
        else:
            additional_values = tuple(tensor_5d(headers[field]) for field in additional_fields)
            return acq_info, additional_values


def write_acqinfo_to_ismrmrd_acquisition_(
    acq_info: AcqInfo,
    ismrmrd_acq: ismrmrd.Acquisition,
    convert_time_stamp: Callable[[float], int] = convert_time_stamp_to_siemens,
) -> ismrmrd.Acquisition:
    """Overwrite ISMRMRD acquisition information for single acquisition.

    Parameters
    ----------
    acq_info
        Acquisition information to write to ISMRMRD acquisition.
    ismrmrd_acq
        ISMRMRD acquisition to write to. Modified in place.
    convert_time_stamp
        Function to convert time stamp to ISMRMRD time stamp.

    Returns
    -------
        ISMRMRD acquisition with updated information.
    """
    if acq_info.shape.numel() != 1:
        raise ValueError('Only single acquisition (single readout for all coils) can be written.')
    ismrmrd_acq.flags = acq_info.flags
    ismrmrd_acq.idx.kspace_encode_step_1 = acq_info.idx.k1
    ismrmrd_acq.idx.kspace_encode_step_2 = acq_info.idx.k2
    ismrmrd_acq.idx.average = acq_info.idx.average
    ismrmrd_acq.idx.slice = acq_info.idx.slice
    ismrmrd_acq.idx.contrast = acq_info.idx.contrast
    ismrmrd_acq.idx.phase = acq_info.idx.phase
    ismrmrd_acq.idx.repetition = acq_info.idx.repetition
    ismrmrd_acq.idx.set = acq_info.idx.set
    ismrmrd_acq.idx.segment = acq_info.idx.segment
    ismrmrd_acq.idx.user = (
        acq_info.idx.user0,
        acq_info.idx.user1,
        acq_info.idx.user2,
        acq_info.idx.user3,
        acq_info.idx.user4,
        acq_info.idx.user5,
        acq_info.idx.user6,
        acq_info.idx.user7,
    )
    # active_channesl, number_of_samples and trajectory_dimensions are read-only and cannot be set
    ismrmrd_acq.patient_table_position = acq_info.patient_table_position[0].apply(m_to_mm).zyx[::-1]  # zyx -> xyz
    directions = acq_info.orientation[0].as_directions()
    ismrmrd_acq.slice_dir = directions[0].zyx[::-1]  # zyx -> xyz
    ismrmrd_acq.phase_dir = directions[1].zyx[::-1]
    ismrmrd_acq.read_dir = directions[2].zyx[::-1]
    ismrmrd_acq.position = acq_info.position[0].apply(m_to_mm).zyx[::-1]
    ismrmrd_acq.sample_time_us = acq_info.sample_time_us
    ismrmrd_acq.user_float = (
        acq_info.user.float0,
        acq_info.user.float1,
        acq_info.user.float2,
        acq_info.user.float3,
        acq_info.user.float4,
        acq_info.user.float5,
        acq_info.user.float6,
        acq_info.user.float7,
    )
    ismrmrd_acq.user_int = (
        acq_info.user.int0,
        acq_info.user.int1,
        acq_info.user.int2,
        acq_info.user.int3,
        acq_info.user.int4,
        acq_info.user.int5,
        acq_info.user.int6,
        acq_info.user.int7,
    )
    ismrmrd_acq.acquisition_time_stamp = convert_time_stamp(acq_info.acquisition_time_stamp.item())
    ismrmrd_acq.physiology_time_stamp = (
        convert_time_stamp(acq_info.physiology_time_stamps.timestamp0.item()),
        convert_time_stamp(acq_info.physiology_time_stamps.timestamp1.item()),
        convert_time_stamp(acq_info.physiology_time_stamps.timestamp2.item()),
    )
    return ismrmrd_acq



"""Proximal Gradient Descent algorithm."""

import math
from collections.abc import Callable
from dataclasses import dataclass

import torch
from typing_extensions import Unpack

from mrpro.algorithms.optimizers.OptimizerStatus import OptimizerStatus
from mrpro.operators import ProximableFunctionalSeparableSum
from mrpro.operators.Functional import ProximableFunctional
from mrpro.operators.Operator import Operator


@dataclass
class PGDStatus(OptimizerStatus):
    """Proximal Gradient Descent callback base class."""

    stepsize: float | torch.Tensor
    objective: Callable[..., torch.Tensor]


def pgd(
    f: Operator[torch.Tensor, tuple[torch.Tensor]] | Operator[Unpack[tuple[torch.Tensor, ...]], tuple[torch.Tensor]],
    g: ProximableFunctional | ProximableFunctionalSeparableSum,
    initial_value: torch.Tensor | tuple[torch.Tensor, ...],
    stepsize: float = 1.0,
    max_iterations: int = 128,
    backtrack_factor: float = 1.0,
    convergent_iterates_variant: bool = False,
    callback: Callable[[PGDStatus], None | bool] | None = None,
) -> tuple[torch.Tensor, ...]:
    r"""Proximal gradient descent algorithm for solving problem :math:`min_x f(x) + g(x)`.

    f is convex, differentiable, and with L-Lipschitz gradient.
    g is convex, possibly non-smooth with computable proximal map.
    g can be a single functional or a `~mrpro.operators.ProximableFunctionalSeparableSum` (e.g. ``g = g1 | g2``).

    For fixed stepsize t, pgd converges globally when :math:`t \in (0, 1/L)`,
    where L is the Lipschitz constant of the gradient of f.
    In applications, f is often of the form :math:`f(x) = 1/2 \|Ax - y\|^2`,
    where :math:`A` is a linear operator.
    In this case, :math:`t \in (0, 1/\|A\|_2^2)` for convergence.
    If no backtracking is used, the fixed stepsize should be given accordingly to the convergence condition.

    Example:
        L1 regularized image reconstruction. Problem formulation: :math:`min_x 1/2 ||Fx - y||^2 + \lambda ||x||_1`,
        with :math:`F` being the Fourier Transform, target denoting the acquired data \in k-space and x \in image space,
        :math:`f(x) = 1/2 \|Fx - y\|^2, g(x) = \lambda \|x\|_1`.
        In this case, :math:`||F^T F|| = 1`. ::
                kspace_data = torch.randn(3, 10, 10 , dtype=torch.complex64)
                fft = FastFourierOp()
                l2 = L2NormSquared(target=kspace_data)
                f = l2 @ fft
                reg_parameter = 0.01
                g = reg_parameter * L1Norm()
                operator_norm = 1.
                stepsize = 0.85 * 1 / operator_norm**2
                initial_value = torch.ones(3,10,10)
                pgd_image_solution = pgd(
                    f=f,
                    g=g,
                    initial_value=initial_value,
                    stepsize=stepsize,
                    max_iterations=200,
                    backtrack_factor=1.0,
                )

    References
    ----------
    .. [BE2009] Beck A, Teboulle M (2009) A Fast Iterative Shrinkage-Thresholding Algorithm for Linear Inverse Problems.
       SIAM Journal on Imaging Sciences
       https://www.ceremade.dauphine.fr/~carlier/FISTA
    .. [CHAM2015] Chambolle A, Dossal C (2015) On the convergence of the iterates of "FISTA".
       Journal of Optimization Theory and Applications
       https://inria.hal.science/hal-01060130v3

    Parameters
    ----------
    f
        convex, differentiable functional
    g
        convex, possibly non-smooth functional with computable proximal map
    initial_value
        initial value for the solution of the algorithm
    stepsize
        stepsize needed in the gradient step, constant if no backtracking is used, otherwise it is the initial stepsize
    max_iterations
        number of iterations
    backtrack_factor
        must be :math:`<=1`. if :math:`<1.`, the backtracking rule for stepsize introduced by [BE2009]_ is used
    convergent_iterates_variant
        by default, the algorithm updates the variable t as originally described in [BE2009]_.
        If set to `True`, the algorithm updates t as suggested by [CHAM2015]_,
        i.e. at iteration :math:`n`, :math:`t_n = \frac{n+a-1}{a}`, with chosen :math:`a=3`.
        This choice ensures the theoretical convergence of solution.
    callback
        function to be called at each iteration. This can be used to monitor the progress of the algorithm.
        If it returns `False`, the algorithm stops at that iteration.

    Returns
    -------
        an approximate solution of the minimization problem
    """
    # check that backtracking factor is in the correct range
    if not 0.0 <= backtrack_factor <= 1.0:
        raise ValueError('Backtracking factor must be in the range [0, 1].')

    backtracking = not math.isclose(backtrack_factor, 1)

    if isinstance(initial_value, torch.Tensor):
        initial_values: tuple[torch.Tensor, ...] = (initial_value,)
    else:
        initial_values = initial_value

    if isinstance(g, ProximableFunctional):
        g_sum = ProximableFunctionalSeparableSum(g)
    else:
        g_sum = g

    x_old = initial_values
    y: tuple[torch.Tensor, ...] = initial_values
    gradient: tuple[torch.Tensor, ...]
    f_y: torch.Tensor
    t_old = 1.0
    grad_and_value_f = torch.func.grad_and_value(
        lambda x: f(*x)[0],
    )
    for iteration in range(max_iterations):
        while stepsize > 1e-30:
            gradient, f_y = grad_and_value_f(y)
            x = g_sum.prox(*[yi - stepsize * gi for yi, gi in zip(y, gradient, strict=True)], sigma=stepsize)

            if not backtracking:
                # no need to check stepsize, continue to next iteration
                break
            difference = tuple(xi - yi for xi, yi in zip(x, y, strict=True))
            quadratic_approx = (
                f_y
                + 1 / (2 * stepsize) * sum(di.abs().square().sum() for di in difference)
                + sum(torch.vdot(gi.flatten(), di.flatten()).real for gi, di in zip(gradient, difference, strict=True))
            )

            (f_x,) = f(*x)

            if f_x <= quadratic_approx:
                # stepsize is ok, continue to next iteration
                break
            stepsize *= backtrack_factor

        else:
            if backtracking:
                raise RuntimeError('After backtracking, the stepsize became to small.')
            else:
                raise RuntimeError('Stepsize to small.')

        if convergent_iterates_variant:
            t = (iteration + 2) / 3
        else:
            t = (1 + math.sqrt(1 + 4 * t_old**2)) / 2

        y = tuple(xi + (t_old - 1.0) / t * (xi - xi_old) for xi, xi_old in zip(x, x_old, strict=True))

        x_old = x
        t_old = t

        if callback is not None:
            continue_iterations = callback(
                PGDStatus(
                    solution=x, iteration_number=iteration, stepsize=stepsize, objective=lambda *x: f(*x)[0] + g(*x)[0]
                )
            )
            if continue_iterations is False:
                break

    return x



"""Jacobian."""

from collections.abc import Callable

import torch

from mrpro.operators.LinearOperator import LinearOperator
from mrpro.operators.Operator import Operator


class Jacobian(LinearOperator):
    """Jacobian of an Operator.

    This operator implements the Jacobian of a (non-linear) operator at a given point :math:`x_0` as a LinearOperator,
    i.e. a linearization of the operator at the point :math:`x_0`.
    """

    def __init__(self, operator: Operator[torch.Tensor, tuple[torch.Tensor]], *x0: torch.Tensor):
        """Initialize the Jacobian operator.

        Parameters
        ----------
        operator
            operator to linearize
        x0
            point at which to linearize the operator
        """
        super().__init__()
        self._vjp: Callable[[tuple[torch.Tensor, ...]], tuple[torch.Tensor, ...]] | None = None
        self._x0: tuple[torch.Tensor, ...] = x0
        self._operator = operator
        self._f_x0: tuple[torch.Tensor, ...] | None = None

    def adjoint(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:  # type:ignore[override]
        """Apply the adjoint operator.

        Parameters
        ----------
        x
            input tensor

        Returns
        -------
            output tensor
        """
        if self._vjp is None:
            self._f_x0, self._vjp = torch.func.vjp(self._operator, *self._x0)
        assert self._vjp is not None  # noqa: S101 (hint for mypy)
        return (self._vjp(x)[0],)

    def __call__(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:  # type:ignore[override]
        """Apply the operator.

        Parameters
        ----------
        x
            input tensor

        Returns
        -------
            output tensor
        """
        return super().__call__(*x)

    def forward(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:  # type:ignore[override]
        """Apply the operator.

        .. note::
            Prefer calling the instance of the Jacobian as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        self._f_x0, jvp = torch.func.jvp(self._operator, self._x0, x)
        return jvp

    @property
    def value_at_x0(self) -> tuple[torch.Tensor, ...]:
        """Evaluation of the operator at the point :math:`x_0`."""
        if self._f_x0 is None:
            self._f_x0 = self._operator(*self._x0)
        assert self._f_x0 is not None  # noqa: S101 (hint for mypy)
        return self._f_x0

    def taylor(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Taylor approximation of the operator.

        Approximate the operator at x by a first order Taylor expansion around :math:`x_0`,
        :math:`f(x_0) + J_f(x_0)(x - x_0)`.

        This is not faster than the forward method of the operator itself, as the calculation of the
        jacobian-vector-product requires the forward pass of the operator to be computed.

        Parameters
        ----------
        x
            input tensor

        Returns
        -------
            Value of the Taylor approximation at x
        """
        delta = tuple(ix - ix0 for ix, ix0 in zip(x, self._x0, strict=False))
        self._f_x0, jvp = torch.func.jvp(self._operator, self._x0, delta)
        assert self._f_x0 is not None  # noqa: S101 (hint for mypy)
        f_x = tuple(ifx0 + ijvp for ifx0, ijvp in zip(self._f_x0, jvp, strict=False))
        return f_x

    def gauss_newton(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Calculate the Gauss-Newton approximation of the Hessian of the operator.

        Returns :math:`J^T J x`, where :math:`J` is the Jacobian of the operator at :math:`x_0`.
        Uses backward and forward automatic differentiation of the operator.

        Parameters
        ----------
        x
            input tensor

        Returns
        -------
            Gauss-Newton approximation of the Hessian applied to x
        """
        return self.adjoint(*self(*x))



"""Density compensation data (DcfData) class."""

from functools import reduce
from typing import TYPE_CHECKING

import torch
from typing_extensions import Self

from mrpro.algorithms.dcf.dcf_voronoi import dcf_1d, dcf_2d3d_voronoi
from mrpro.data.Dataclass import Dataclass
from mrpro.data.KTrajectory import KTrajectory
from mrpro.utils import smap

if TYPE_CHECKING:
    from mrpro.operators.DensityCompensationOp import DensityCompensationOp


class DcfData(Dataclass):
    """Density compensation data (DcfData) class."""

    data: torch.Tensor
    """Density compensation values. Shape `(... other, coils, k2, k1, k0)`"""

    @classmethod
    def from_traj_voronoi(cls, traj: KTrajectory) -> Self:
        """Calculate dcf using voronoi approach for 2D or 3D trajectories.

        Parameters
        ----------
        traj
            Trajectory to calculate the density compensation for. Can be broadcasted or dense.
        """
        dcfs = []

        ks = [traj.kz, traj.ky, traj.kx]
        spatial_dims = (-3, -2, -1)
        ks_needing_voronoi = set()
        for dim in spatial_dims:
            non_singleton_ks = [ax for ax in ks if ax.shape[dim] != 1]
            if len(non_singleton_ks) == 1:
                # Found a dimension with only one non-singleton axes in ks
                # --> Can handle this as a 1D trajectory
                dcfs.append(smap(dcf_1d, non_singleton_ks.pop(), (dim,)))
            elif len(non_singleton_ks) > 0:
                # More than one of the ks is non-singleton
                # --> A full dimension needing voronoi
                ks_needing_voronoi |= set(non_singleton_ks)
            else:
                # A dimension in which each of ks is singleton
                # --> Don't need to do anything
                pass

        if ks_needing_voronoi:
            # Handle full dimensions needing voronoi
            dcfs.append(smap(dcf_2d3d_voronoi, torch.stack(torch.broadcast_tensors(*ks_needing_voronoi), -4), 4))

        if dcfs:
            # Multiply all dcfs together
            dcf = reduce(torch.mul, dcfs)
        else:
            # Edgecase: Only singleton spatial dimensions
            dcf = torch.ones(*traj.shape[-3:], 1, 1, 1, device=traj.kx.device)

        return cls(data=dcf)

    def as_operator(self) -> 'DensityCompensationOp':
        """Create a density compensation operator using a copy of the DCF."""
        from mrpro.operators.DensityCompensationOp import DensityCompensationOp

        return DensityCompensationOp(self.data.clone())



"""Efficient sliding window implementation for tensors."""

from collections.abc import Sequence

import torch


def sliding_window(
    x: torch.Tensor,
    window_shape: int | Sequence[int],
    dim: int | Sequence[int] | None = None,
    stride: int | Sequence[int] | None = 1,
    dilation: int | Sequence[int] = 1,
) -> torch.Tensor:
    """Create a sliding window view into a tensor.

    Returns a view into tensor x where new dim representing the number of windows
    are added at the front, and the original dim involved in the sliding operation
    are replaced by window dimensions.

    Example:

        Input shape (D1, D2, D3, D4, D5), dim=(1, 3), window_shape=(k2, k4)
        Output shape: (n_windows_2, n_windows_4, D1, k2, D3, k4, D5)

    Parameters
    ----------
    x
        Input tensor to create sliding windows from
    window_shape
        Size of window over each dimension
    dim
        Dimension(s) to apply to. If None, applies to all dimensions.
    stride
        Stride of the sliding window. If None, equals window_shape.
    dilation
        Spacing between window elements for each dimension.

    Returns
    -------
        A view of the tensor with window dimensions at the front and
        original sliding dim replaced by kernel dimensions.
    """
    ndim = x.ndim

    if dim is None:
        dim = tuple(range(ndim))
    elif isinstance(dim, int):
        dim = (dim % ndim,)
    else:
        dim = tuple(ax % ndim for ax in dim)
        if len(set(dim)) != len(dim):
            raise ValueError('Duplicate values in axis are not allowed')

    n_dim = len(dim)
    window_shape_ = (window_shape,) * n_dim if isinstance(window_shape, int) else tuple(window_shape)
    stride_ = window_shape_ if stride is None else ((stride,) * n_dim if isinstance(stride, int) else tuple(stride))
    dilation_ = (dilation,) * n_dim if isinstance(dilation, int) else tuple(dilation)

    if any(len(param) != n_dim for param in [window_shape_, stride_, dilation_]):
        raise ValueError(f'Length mismatch: window_shape, stride, and dilation must all have length {n_dim}')

    if any(w <= 0 for w in window_shape_):
        raise ValueError('window_shape must be positive')
    if any(s <= 0 for s in stride_):
        raise ValueError('stride must be positive')
    if any(d <= 0 for d in dilation_):
        raise ValueError('dilation must be positive')

    effective_sizes = [(d * (w - 1) + 1) for w, d in zip(window_shape_, dilation_, strict=False)]
    axis_to_idx = {ax: i for i, ax in enumerate(dim)}
    out_shape = []
    for i, ax in enumerate(dim):
        n_win = (x.shape[ax] - effective_sizes[i]) // stride_[i] + 1
        if n_win <= 0:
            axis_size = x.shape[ax]
            raise ValueError(
                f'Dimension {ax} with size {axis_size} is too small for '
                f'window_size={window_shape_[i]}, dilation={dilation_[i]}, stride={stride_[i]}'
            )
        out_shape.append(n_win)
    x_stride = x.stride()
    out_stride = [x_stride[ax] * st for ax, st in zip(dim, stride_, strict=False)]
    for i in range(len(x.shape)):
        if i in axis_to_idx:
            idx = axis_to_idx[i]
            out_shape.append(window_shape_[idx])
            out_stride.append(x_stride[i] * dilation_[idx])
        else:
            out_shape.append(x.shape[i])
            out_stride.append(x_stride[i])

    return x.as_strided(size=out_shape, stride=out_stride)



"""MR raw data / k-space data header dataclass."""

import dataclasses
import datetime as dt
import warnings
from collections.abc import Mapping
from dataclasses import field
from typing import TYPE_CHECKING

import ismrmrd.xsd.ismrmrdschema.ismrmrd as ismrmrdschema
import torch
from typing_extensions import Self

from mrpro.data import enums
from mrpro.data.AcqInfo import AcqInfo
from mrpro.data.Dataclass import Dataclass
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator
from mrpro.utils.unit_conversion import (
    deg_to_rad,
    lamor_frequency_to_magnetic_field,
    m_to_mm,
    mm_to_m,
    ms_to_s,
    rad_to_deg,
    s_to_ms,
)

if TYPE_CHECKING:
    # avoid circular imports by importing only when type checking
    from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator

UNKNOWN = 'unknown'


class KHeader(Dataclass):
    """MR raw data header.

    All information that is not covered by the dataclass is stored in
    the misc dict. Our code shall not rely on this information, and it is
    not guaranteed to be present. Also, the information in the misc dict
    is not guaranteed to be correct or tested.
    """

    recon_matrix: SpatialDimension[int]
    """Dimensions of the reconstruction matrix."""

    encoding_matrix: SpatialDimension[int]
    """Dimensions of the encoded k-space matrix."""

    recon_fov: SpatialDimension[float]
    """Field-of-view of the reconstructed image [m]."""

    encoding_fov: SpatialDimension[float]
    """Field of view of the image encoded by the k-space trajectory [m]."""

    acq_info: AcqInfo = field(default_factory=AcqInfo)
    """Information of the acquisitions (i.e. readout lines)."""

    trajectory: KTrajectoryCalculator | None = None
    """Function to calculate the k-space trajectory."""

    lamor_frequency_proton: float | None = None
    """Lamor frequency of hydrogen nuclei [Hz]."""

    datetime: dt.datetime | None = None
    """Date and time of acquisition."""

    te: list[float] | torch.Tensor = field(default_factory=list)
    """Echo time [s]."""

    ti: list[float] | torch.Tensor = field(default_factory=list)
    """Inversion time [s]."""

    fa: list[float] | torch.Tensor = field(default_factory=list)
    """Flip angle [rad]."""

    tr: list[float] | torch.Tensor = field(default_factory=list)
    """Repetition time [s]."""

    echo_spacing: list[float] | torch.Tensor = field(default_factory=list)
    """Echo spacing [s]."""

    echo_train_length: int = 1
    """Number of echoes in a multi-echo acquisition."""

    sequence_type: str = UNKNOWN
    """Type of sequence."""

    model: str = UNKNOWN
    """Scanner model."""

    vendor: str = UNKNOWN
    """Scanner vendor."""

    protocol_name: str = UNKNOWN
    """Name of the acquisition protocol."""

    trajectory_type: enums.TrajectoryType = enums.TrajectoryType.OTHER
    """Type of trajectory."""

    measurement_id: str = UNKNOWN
    """Measurement ID."""

    patient_name: str = UNKNOWN
    """Name of the patient."""

    _misc: dict = field(default_factory=dict)
    """Dictionary with miscellaneous parameters. These parameters are for information purposes only. Reconstruction
    algorithms should not rely on them."""

    @property
    def fa_degree(self) -> torch.Tensor | list[float]:
        """Flip angle in degree."""
        if isinstance(self.fa, torch.Tensor):
            return torch.rad2deg(self.fa)
        elif not len(self.fa):
            warnings.warn('Flip angle is not defined.', stacklevel=1)
            return []
        else:
            return torch.rad2deg(torch.as_tensor(self.fa)).tolist()

    @classmethod
    def from_ismrmrd(
        cls,
        header: ismrmrdschema.ismrmrdHeader,
        acq_info: AcqInfo,
        defaults: dict | None = None,
        overwrite: Mapping[str, object] | None = None,
        encoding_number: int = 0,
    ) -> Self:
        """Create an Header from ISMRMRD Data.

        Parameters
        ----------
        header
            ISMRMRD header
        acq_info
            acquisition information
        defaults
            dictionary of values to be used if information is missing in header
        overwrite
            dictionary of values to be used independent of header
        encoding_number
            as ismrmrdHeader can contain multiple encodings, selects which to consider
        """
        if not 0 <= encoding_number < len(header.encoding):
            raise ValueError(f'encoding_number must be between 0 and {len(header.encoding)}')

        enc: ismrmrdschema.encodingType = header.encoding[encoding_number]

        # These are guaranteed to exist
        parameters = {
            'lamor_frequency_proton': header.experimentalConditions.H1resonanceFrequency_Hz,
            'acq_info': acq_info,
        }

        if defaults is not None:
            parameters.update(defaults)

        if header.sequenceParameters is not None:
            if header.sequenceParameters.TR:
                parameters['tr'] = ms_to_s(header.sequenceParameters.TR)
            if header.sequenceParameters.TE:
                parameters['te'] = ms_to_s(header.sequenceParameters.TE)
            if header.sequenceParameters.TI:
                parameters['ti'] = ms_to_s(header.sequenceParameters.TI)
            if header.sequenceParameters.flipAngle_deg:
                parameters['fa'] = deg_to_rad(header.sequenceParameters.flipAngle_deg)
            if header.sequenceParameters.echo_spacing:
                parameters['echo_spacing'] = ms_to_s(header.sequenceParameters.echo_spacing)

            if header.sequenceParameters.sequence_type is not None:
                parameters['sequence_type'] = header.sequenceParameters.sequence_type

        if enc.reconSpace is not None:
            parameters['recon_fov'] = SpatialDimension[float].from_xyz(enc.reconSpace.fieldOfView_mm).apply_(mm_to_m)
            parameters['recon_matrix'] = SpatialDimension[int].from_xyz(enc.reconSpace.matrixSize)

        if enc.encodedSpace is not None:
            parameters['encoding_fov'] = (
                SpatialDimension[float].from_xyz(enc.encodedSpace.fieldOfView_mm).apply_(mm_to_m)
            )
            parameters['encoding_matrix'] = SpatialDimension[int].from_xyz(enc.encodedSpace.matrixSize)

        if enc.echoTrainLength is not None:
            parameters['echo_train_length'] = enc.echoTrainLength

        if enc.trajectory is not None:
            parameters['trajectory_type'] = enums.TrajectoryType(enc.trajectory.value)

        # Either use the series or study time if available
        if header.measurementInformation is not None and header.measurementInformation.seriesTime is not None:
            time = header.measurementInformation.seriesTime.to_time()
        elif header.studyInformation is not None and header.studyInformation.studyTime is not None:
            time = header.studyInformation.studyTime.to_time()
        else:  # if no time is given, set to 00:00:00
            time = dt.time()
        if header.measurementInformation is not None and header.measurementInformation.seriesDate is not None:
            date = header.measurementInformation.seriesDate.to_date()
            parameters['datetime'] = dt.datetime.combine(date, time)
        elif header.studyInformation is not None and header.studyInformation.studyDate is not None:
            date = header.studyInformation.studyDate.to_date()
            parameters['datetime'] = dt.datetime.combine(date, time)

        if header.subjectInformation is not None and header.subjectInformation.patientName is not None:
            parameters['patient_name'] = header.subjectInformation.patientName

        if header.measurementInformation is not None:
            if header.measurementInformation.measurementID is not None:
                parameters['measurement_id'] = header.measurementInformation.measurementID

            if header.measurementInformation.protocolName is not None:
                parameters['protocol_name'] = header.measurementInformation.protocolName

        if header.acquisitionSystemInformation is not None:
            if header.acquisitionSystemInformation.systemVendor is not None:
                parameters['vendor'] = header.acquisitionSystemInformation.systemVendor

            if header.acquisitionSystemInformation.systemModel is not None:
                parameters['model'] = header.acquisitionSystemInformation.systemModel

        # Dump everything into misc
        parameters['_misc'] = dataclasses.asdict(header)

        if overwrite is not None:
            parameters.update(overwrite)

        try:
            instance = cls(**parameters)
        except TypeError:
            missing = [
                f.name
                for f in dataclasses.fields(cls)
                if f.name not in parameters
                and (f.default == dataclasses.MISSING and f.default_factory == dataclasses.MISSING)
            ]
            raise ValueError(
                f'Could not create Header. Missing parameters: {missing}\n'
                'Consider setting them via the defaults dictionary',
            ) from None
        return instance

    def to_ismrmrd(self) -> ismrmrdschema.ismrmrdHeader:
        """Create ISMRMRD header from KHeader.

        All the parameters from the ISMRMRD header of the raw file used to create the kheader are saved in the
        dictionary kheader['misc']. We first fill the information from this dictionary into the ISMRMRD header and then
        we update the values based on the attributes of the KHeader.

        Returns
        -------
            ISMRMRD header
        """
        header = ismrmrdschema.ismrmrdHeader()

        # Experimental conditions
        exp = ismrmrdschema.experimentalConditionsType()
        exp.H1resonanceFrequency_Hz = self.lamor_frequency_proton
        header.experimentalConditions = exp

        # Subject information
        subj = ismrmrdschema.subjectInformationType()
        subj.patientName = self.patient_name
        header.subjectInformation = subj

        # Measurement information
        meas = ismrmrdschema.measurementInformationType()
        meas.protocolName = self.protocol_name
        meas.measurementID = self.measurement_id
        if self.datetime is not None:
            meas.seriesTime = str(self.datetime.time())
            meas.seriesDate = str(self.datetime.date())
        header.measurementInformation = meas

        # Acquisition system information
        sys = ismrmrdschema.acquisitionSystemInformationType()
        sys.systemModel = self.model
        sys.systemVendor = self.vendor
        if self.lamor_frequency_proton:
            sys.systemFieldStrength_T = lamor_frequency_to_magnetic_field(self.lamor_frequency_proton)
        header.acquisitionSystemInformation = sys

        # Sequence information
        seq = ismrmrdschema.sequenceParametersType()
        seq.TR = s_to_ms(self.tr).tolist() if isinstance(self.tr, torch.Tensor) else s_to_ms(self.tr)
        seq.TE = s_to_ms(self.te).tolist() if isinstance(self.te, torch.Tensor) else s_to_ms(self.te)
        seq.TI = s_to_ms(self.ti).tolist() if isinstance(self.ti, torch.Tensor) else s_to_ms(self.ti)
        seq.flipAngle_deg = rad_to_deg(self.fa).tolist() if isinstance(self.fa, torch.Tensor) else rad_to_deg(self.fa)
        seq.echo_spacing = (
            s_to_ms(self.echo_spacing).tolist()
            if isinstance(self.echo_spacing, torch.Tensor)
            else s_to_ms(self.echo_spacing)
        )
        seq.sequence_type = self.sequence_type
        header.sequenceParameters = seq

        # Encoding
        encoding = ismrmrdschema.encodingType()

        # Encoded space
        encoding_space = ismrmrdschema.encodingSpaceType()
        encoding_fov = ismrmrdschema.fieldOfViewMm(
            m_to_mm(self.encoding_fov.x), m_to_mm(self.encoding_fov.y), m_to_mm(self.encoding_fov.z)
        )
        encoding_matrix = ismrmrdschema.matrixSizeType(
            self.encoding_matrix.x, self.encoding_matrix.y, self.encoding_matrix.z
        )
        encoding_space.matrixSize = encoding_matrix
        encoding_space.fieldOfView_mm = encoding_fov

        # Recon space
        recon_space = ismrmrdschema.encodingSpaceType()
        recon_fov = ismrmrdschema.fieldOfViewMm(
            m_to_mm(self.recon_fov.x), m_to_mm(self.recon_fov.y), m_to_mm(self.recon_fov.z)
        )
        recon_matrix = ismrmrdschema.matrixSizeType(self.recon_matrix.x, self.recon_matrix.y, self.recon_matrix.z)
        recon_space.matrixSize = recon_matrix
        recon_space.fieldOfView_mm = recon_fov

        # Set encoded and recon spaces
        encoding.encodedSpace = encoding_space
        encoding.reconSpace = recon_space
        header.encoding.append(encoding)

        return header



"""Dictionary Matching Operator."""

from collections.abc import Callable
from typing import cast

import einops
import torch
from typing_extensions import Self, TypeVarTuple, Unpack

from mrpro.operators.Operator import Operator

Tin = TypeVarTuple('Tin')


class DictionaryMatchOp(Operator[torch.Tensor, tuple[Unpack[Tin]]]):
    r"""Dictionary Matching Operator.

    This operator can be used for dictionary matching, for example in
    magnetic ressonance fingerprinting.

    It performs absolute normalized dot product matching between a dictionary of signals,
    i.e. find the entry :math:`d^*` in the dictionary maximizing
    :math:`\left|\frac{d}{\|d\|} \cdot \frac{y}{\|y\|}\right|` and returns the
    associated signal model parameters :math:`x` generating the matching signal :math:`d^*=d(x)`.

    At initialization, a signal model needs to be provided.
    Afterwards `append` with different `x` values should be called to add entries to the dictionary.
    This operator then calculates for each `x` value the signal returned by the model.
    To perform a match, use `__call__` and supply some `y` values. The operator will then perform the
    dot product matching and return the associated `x` values.
    """

    def __init__(
        self,
        generating_function: Callable[[Unpack[Tin]], tuple[torch.Tensor,]],
        index_of_scaling_parameter: int | None = None,
    ):
        """Initialize DictionaryMatchOp.

        Parameters
        ----------
        generating_function
            signal model that takes n inputs and returns a signal y.
        index_of_scaling_parameter
            Normalized dot product matching is insensitive to overall signal scaling.
            A scaling factor (e.g. the equilibrium magnetization `m0` in `~mrpro.operators.models.InversionRecovery`)
            is calculated after the dictionary matching if `index_of_scaling_parameter` is not `None`.
            `index_of_scaling_parameter` should set to the index of the scaling parameter in the signal model.

            Example:
                For ~mrpro.operators.models.InversionRecovery the parameters are ``[m0, t1]`` and therefore
                `index_of_scaling_parameter` should be set to 0. The operator will then return `t1` estimated
                via dictionary matching and `m0` via a post-processing step.
                If `index_of_scaling_parameter` is None, the value returned for `m0` will be meaningless.
        """
        super().__init__()
        self._f = generating_function
        self.x: list[torch.Tensor] = []
        self.y = torch.tensor([])
        self._index_of_scaling_parameter = index_of_scaling_parameter
        self.inverse_norm_y = None if index_of_scaling_parameter is None else torch.tensor([])

    def append(self, *x: Unpack[Tin]) -> Self:
        """Append `x` values to the dictionary.

        Parameters
        ----------
        x
            points where the signal model will be evaluated. For signal models
            with n inputs, n tensors should be provided. Broadcasting is supported.

        Returns
        -------
            Self

        """
        if self._index_of_scaling_parameter is not None:
            scaling_position = self._index_of_scaling_parameter % len(x)
            # replace the scaling argument with 1 in call
            (y,) = self._f(*x[:scaling_position], torch.tensor(1), *x[scaling_position + 1 :])  # type: ignore[call-arg]
            # but drop it in the dictionary
            x_list = [x.flatten() for x in torch.broadcast_tensors(*x[:scaling_position], *x[scaling_position + 1 :])]
        else:
            (y,) = self._f(*x)
            x_list = [x.flatten() for x in torch.broadcast_tensors(*x)]

        y = y.flatten(start_dim=1)
        inverse_norm_y = torch.linalg.norm(y, dim=0).reciprocal()
        y = y * inverse_norm_y

        if not self.x:
            self.x = x_list
            self.y = y
            if self.inverse_norm_y is not None:
                self.inverse_norm_y = inverse_norm_y
            return self

        self.x = [torch.cat((old, new)) for old, new in zip(self.x, x_list, strict=True)]
        self.y = torch.cat((self.y, y), dim=-1)
        if self.inverse_norm_y is not None:
            self.inverse_norm_y = torch.cat((self.inverse_norm_y, inverse_norm_y))
        return self

    def __call__(self, input_signal: torch.Tensor) -> tuple[Unpack[Tin]]:
        """Perform dot-product matching.

        Given an input signal (or batch of signals), this method finds the entry in the
        precomputed dictionary that has the highest absolute normalized dot product similarity.
        It then returns the set of parameters `x` that generated this dictionary entry.

        If `index_of_scaling_parameter` was set during initialization, the corresponding
        scaling parameter is estimated and inserted into the returned tuple of parameters.

        Parameters
        ----------
        input_signal
            Input signal(s) to match against the dictionary.
            Expected shape is `(m, ...)`, where `m` is the signal dimension
            (e.g., number of time points) and `(...)` are batch dimensions.

        Returns
        -------
            A tuple of tensors representing the parameters `x` from the dictionary
            that best matched the input signal(s). Each tensor in the tuple corresponds
            to a parameter, and their shapes will match the batch dimensions of `input_signal`.
        """
        return super().__call__(input_signal)

    def forward(self, input_signal: torch.Tensor) -> tuple[Unpack[Tin]]:
        """Apply forward of DictionaryMatchOp.

        .. note::
            Prefer calling the instance of the DictionaryMatchOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if not self.x:
            raise KeyError('No keys in the dictionary. Please first add some x values using `append`.')

        # This avoids unnecessary copies mixed domain cases
        similarity = einops.einsum(input_signal.real, self.y.real, 'm ..., m idx  -> idx ...').square()
        if self.y.is_complex():
            similarity += einops.einsum(input_signal.real, self.y.imag, 'm ..., m idx  -> idx ...').square()
        if input_signal.is_complex():
            similarity += einops.einsum(input_signal.imag, self.y.real, 'm ..., m idx  -> idx ...').square()
        if self.y.is_complex() and input_signal.is_complex():
            similarity += einops.einsum(input_signal.imag, self.y.imag, 'm ..., m idx  -> idx ...').square()

        idx = similarity.argmax(dim=0)
        match = [x[idx] for x in self.x]

        if self._index_of_scaling_parameter is not None and self.inverse_norm_y is not None:
            # replace the scaling argument with the correct scaling factor
            scale = (self.y[:, idx].conj() * input_signal).sum(0) * self.inverse_norm_y[idx]
            match.insert(self._index_of_scaling_parameter, scale)

        return cast(tuple[Unpack[Tin]], tuple(match))



"""Radial phase encoding (RPE) trajectory class with sunflower pattern."""

import numpy as np
import torch

from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator
from mrpro.utils.reshape import unsqueeze_tensors_left

GOLDEN_RATIO = 0.5 * (5**0.5 + 1)


class KTrajectorySunflowerGoldenRpe(KTrajectoryCalculator):
    """Radial phase encoding trajectory with a sunflower pattern."""

    def __init__(self) -> None:
        """Initialize KTrajectorySunflowerGoldenRpe."""
        self.angle = torch.pi * 0.618034

    def _apply_sunflower_shift_between_rpe_lines(
        self,
        radial: torch.Tensor,
        angles: torch.Tensor,
        k2_idx: torch.Tensor,
    ) -> torch.Tensor:
        """Shift radial phase encoding lines relative to each other.

        The shifts are applied to create a sunflower pattern of k-space points in the ky-kz phase encoding plane.

        Parameters
        ----------
        radial
            position along radial direction
        angles
            angle of spokes
        k2_idx
            indices in k2

        Returns
        -------
            shifted radial positions
        """
        angles = angles.flatten()
        _, indices = np.unique(angles, return_index=True)
        shift_idx = np.argsort(indices)
        k2_idx, radial = torch.broadcast_tensors(k2_idx, radial)
        radial = radial.contiguous()
        for ind, shift in enumerate(shift_idx):
            radial[k2_idx == ind] += ((shift * GOLDEN_RATIO) % 1) - 0.5
        return radial

    def __call__(
        self,
        *,
        n_k0: int,
        k0_center: int | torch.Tensor,
        k1_idx: torch.Tensor,
        k1_center: int | torch.Tensor,
        k2_idx: torch.Tensor,
        reversed_readout_mask: torch.Tensor | None = None,
        **_,
    ) -> KTrajectory:
        """Calculate radial phase encoding trajectory for given header information.

        Parameters
        ----------
        n_k0
            number of samples in k0
        k0_center
            position of k-space center in k0
        k1_idx
            indices of k1 (radial)
        k1_center
            position of k-space center in k1
        k2_idx
            indices of k2 (angle)
        reversed_readout_mask
            boolean tensor indicating reversed readout

        Returns
        -------
            radial phase encoding trajectory for given KHeader
        """
        angles = (k2_idx * self.angle) % torch.pi

        radial = (k1_idx - k1_center).to(torch.float32)
        radial = self._apply_sunflower_shift_between_rpe_lines(radial, angles, k2_idx)
        # Asymmetric k-space point is used to obtain a self-navigator signal, thus should be in k-space center
        radial[(k1_idx == 0).broadcast_to(radial.shape)] = 0

        kz = radial * torch.sin(angles)
        ky = radial * torch.cos(angles)
        kx = self._readout(n_k0, k0_center, reversed_readout_mask=reversed_readout_mask)
        kz, ky, kx = unsqueeze_tensors_left(kz, ky, kx, ndim=5)

        return KTrajectory(kz, ky, kx)



"""Fourier Operator."""

from collections.abc import Sequence
from functools import cached_property

import torch
from typing_extensions import Self

from mrpro.data.enums import TrajType
from mrpro.data.KData import KData
from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.operators.CartesianSamplingOp import CartesianSamplingOp
from mrpro.operators.FastFourierOp import FastFourierOp
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.operators.NonUniformFastFourierOp import NonUniformFastFourierOp


class FourierOp(LinearOperator, adjoint_as_backward=True):
    """Fourier Operator class.

    This is the recommended operator for all Fourier transformations.
    It auto-detects if a non-uniform or regular fast Fourier transformation is required.
    For Cartesian data on a regular grid, the data is sorted and a FFT is used.
    For non-Cartesian data, a NUFFT with regridding is used.
    It also includes padding/cropping to the reconstruction matrix size.

    The operator can directly be constructed from a `~mrpro.data.KData` object to match its
    trajectory and header information, see `FourierOp.from_kdata`.

    """

    def __init__(
        self,
        recon_matrix: SpatialDimension[int],
        encoding_matrix: SpatialDimension[int],
        traj: KTrajectory,
    ) -> None:
        """Initialize Fourier Operator.

        Parameters
        ----------
        recon_matrix
            dimension of the reconstructed image
        encoding_matrix
            dimension of the encoded k-space
        traj
            the k-space trajectories where the frequencies are sampled
        """
        super().__init__()

        def get_spatial_dims(spatial_dims: SpatialDimension, dims: Sequence[int]):
            return [
                s
                for s, i in zip((spatial_dims.z, spatial_dims.y, spatial_dims.x), (-3, -2, -1), strict=True)
                if i in dims
            ]

        self._ignore_dims, self._fft_dims, self._nufft_dims = [], [], []
        for dim, type_ in zip((-3, -2, -1), traj.type_along_kzyx, strict=True):
            if type_ & TrajType.SINGLEVALUE:
                # dimension which do not require any transform
                self._ignore_dims.append(dim)
            elif type_ & TrajType.ONGRID:
                self._fft_dims.append(dim)
            else:
                self._nufft_dims.append(dim)

        if self._fft_dims:
            self._fast_fourier_op: FastFourierOp | None = FastFourierOp(
                dim=tuple(self._fft_dims),
                recon_matrix=get_spatial_dims(recon_matrix, self._fft_dims),
                encoding_matrix=get_spatial_dims(encoding_matrix, self._fft_dims),
            )
            self._cart_sampling_op: CartesianSamplingOp | None = CartesianSamplingOp(
                encoding_matrix=encoding_matrix, traj=traj
            )
        else:
            self._fast_fourier_op = None
            self._cart_sampling_op = None

        # Find dimensions which require NUFFT
        if self._nufft_dims:
            fft_dims_k210 = [
                dim
                for dim in (-3, -2, -1)
                if (traj.type_along_k210[dim] & TrajType.ONGRID)
                and not (traj.type_along_k210[dim] & TrajType.SINGLEVALUE)
            ]
            if self._fft_dims != fft_dims_k210:
                raise NotImplementedError(
                    'If both FFT and NUFFT dims are present, Cartesian FFT dims need to be aligned with the '
                    'k-space dimension, i.e. kx along k0, ky along k1 and kz along k2.',
                )

            self._non_uniform_fast_fourier_op: NonUniformFastFourierOp | None = NonUniformFastFourierOp(
                direction=tuple(self._nufft_dims),  # type: ignore[arg-type]
                recon_matrix=get_spatial_dims(recon_matrix, self._nufft_dims),
                encoding_matrix=get_spatial_dims(encoding_matrix, self._nufft_dims),
                traj=traj,
            )
        else:
            self._non_uniform_fast_fourier_op = None

        self._trajectory_info = repr(traj)

    @classmethod
    def from_kdata(cls, kdata: KData, recon_shape: SpatialDimension[int] | None = None) -> Self:
        """Create an instance of FourierOp from kdata with default settings.

        Parameters
        ----------
        kdata
            k-space data
        recon_shape
            Dimension of the reconstructed image. Defaults to `KData.header.recon_matrix`.
        """
        return cls(
            recon_matrix=kdata.header.recon_matrix if recon_shape is None else recon_shape,
            encoding_matrix=kdata.header.encoding_matrix,
            traj=kdata.traj,
        )

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward Fourier operation (image to k-space).

        This operator maps coil image data to coil k-space data.
        Depending on the trajectory and dimensions, it may involve NUFFT, FFT,
        and Cartesian sampling operations.

        Parameters
        ----------
        x
            Coil image data, typically with shape `(... coils z y x)`.

        Returns
        -------
            Coil k-space data, typically with shape `(... coils k2 k1 k0)`.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of FourierOp.

        .. note::
            Prefer calling the instance of the FourierOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        # NUFFT Type 2 followed by FFT
        if self._non_uniform_fast_fourier_op:
            (x,) = self._non_uniform_fast_fourier_op(x)

        if self._fast_fourier_op and self._cart_sampling_op:
            (x,) = self._cart_sampling_op(self._fast_fourier_op(x)[0])
        return (x,)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply adjoint Fourier operation (k-space to image).

        This operator maps coil k-space data to coil image data.
        It is the adjoint of the forward operation and involves corresponding
        adjoint NUFFT, FFT, and Cartesian sampling operations.

        Parameters
        ----------
        x
            Coil k-space data, typically with shape `(... coils k2 k1 k0)`.

        Returns
        -------
            Coil image data, typically with shape `(... coils z y x)`.
        """
        # FFT followed by NUFFT Type 1
        if self._fast_fourier_op and self._cart_sampling_op:
            (x,) = self._fast_fourier_op.adjoint(self._cart_sampling_op.adjoint(x)[0])

        if self._non_uniform_fast_fourier_op:
            (x,) = self._non_uniform_fast_fourier_op.adjoint(x)
        return (x,)

    @cached_property
    def gram(self) -> LinearOperator:
        """Return the gram operator."""
        try:
            return FourierGramOp(self)
        except NotImplementedError:
            return self.H @ self

    def __repr__(self) -> str:
        """Representation method for Fourier Operator."""
        string = ''
        device_omega = None
        device_cart = None

        if self._nufft_dims and self._non_uniform_fast_fourier_op:
            nufftop = self._non_uniform_fast_fourier_op
            string += f'\n{nufftop}\n'
            device_omega = nufftop._omega.device if nufftop._omega.device is not None else None
        if self._fft_dims:
            string += f'\n{self._fast_fourier_op}\n\n{self._cart_sampling_op}\n'
            device_cart = self._cart_sampling_op._fft_idx.device if self._cart_sampling_op is not None else None

        if device_omega and device_cart:
            device = device_omega if device_omega == device_cart else 'Different devices'
        else:
            device = device_omega or device_cart or 'None'

        out = f'{type(self).__name__} on device: {device}\n{string}\n{self._trajectory_info}'
        return out


class FourierGramOp(LinearOperator):
    """Gram operator for the Fourier operator.

    Implements the adjoint of the forward operator of the Fourier operator, i.e. the gram operator
    `F.H@F`.

    Uses a convolution, implemented as multiplication in Fourier space, to calculate the gram operator
    for the toeplitz NUFFT operator.

    Uses a multiplication with a binary mask in Fourier space to calculate the gram operator for
    the Cartesian FFT operator

    This Operator is only used internally and should not be used directly.
    Instead, consider using the py:func:`~FourierOp.gram` property of py:class:`FourierOp`.
    """

    _kernel: torch.Tensor | None

    def __init__(self, fourier_op: FourierOp) -> None:
        """Initialize the gram operator.

        If density compensation weights are provided, they the operator
        F.H@dcf@F is calculated.

        Parameters
        ----------
        fourier_op
            the Fourier operator to calculate the gram operator for

        """
        super().__init__()

        if fourier_op._fast_fourier_op and fourier_op._cart_sampling_op:
            sampling_gram = fourier_op._cart_sampling_op.gram
            if (
                fourier_op._non_uniform_fast_fourier_op
                and sampling_gram.mask is not None
                and any(sampling_gram.mask.shape[d] != 1 for d in fourier_op._nufft_dims)
            ):
                raise NotImplementedError(
                    'FourierGramOp does not support non-uniform FFTs combined with Cartesian sampling '
                    'that differs along the NUFFT dimensions.'
                )
            self.fast_fourier_gram: None | LinearOperator = (
                fourier_op._fast_fourier_op.H @ fourier_op._cart_sampling_op.gram @ fourier_op._fast_fourier_op
            )
        else:
            self.fast_fourier_gram = None

        if fourier_op._non_uniform_fast_fourier_op:
            self.nufft_gram: None | LinearOperator = fourier_op._non_uniform_fast_fourier_op.gram
        else:
            self.nufft_gram = None

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the Gram operator of the FourierOp (F.H @ F).

        This operation applies the composition of the adjoint Fourier operator
        and the forward Fourier operator. It may involve Gram operators
        of NUFFT and/or FFT components.

        Parameters
        ----------
        x
            Input tensor, typically image-space data with shape `(..., coils, z, y, x)`.

        Returns
        -------
            Output tensor, typically image-space data after F.H @ F has been applied.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of FourierGramOp.

        .. note::
            Prefer calling the instance of the FourierGramOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if self.nufft_gram:
            (x,) = self.nufft_gram(x)

        if self.fast_fourier_gram:
            (x,) = self.fast_fourier_gram(x)
        return (x,)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the adjoint of the Gram operator.

        Since the Gram operator (F.H @ F) is self-adjoint, this method
        calls the forward operation.

        Parameters
        ----------
        x
            Input tensor, typically image-space data with shape `(..., coils, z, y, x)`.
            Note: The original docstring mentioned k-space shape, but for a self-adjoint
            image-to-image Gram operator, the input to adjoint should match input to forward.

        Returns
        -------
            Output tensor, same as the forward operation.
        """
        return self.forward(x)

    @property
    def H(self) -> Self:  # noqa: N802
        """Adjoint operator of the gram operator."""
        return self



"""L2 Squared Norm."""

import torch

from mrpro.operators.Functional import ElementaryProximableFunctional, throw_if_negative_or_complex


class L2NormSquared(ElementaryProximableFunctional):
    r"""Functional class for the squared L2 Norm.

    This implements the functional given by
    :math:`f: C^N \rightarrow [0, \infty), x \rightarrow \| W (x-b)\|_2^2`,
    where :math:`W` is either a scalar or tensor that corresponds to a (block-) diagonal operator
    that is applied to the input. This is, for example, useful for non-Cartesian MRI
    reconstruction when using a density-compensation function for k-space pre-conditioning,
    for masking of image data, or for spatially varying regularization weights.

    In most cases, consider setting `divide_by_n` to `True` to be independent of input size.
    Alternatively, the functional :class:`mrpro.operators.functionals.MSE` can be used.
    The norm is computed along the dimensions given at initialization, all other dimensions are
    considered batch dimensions.
    """

    def __call__(
        self,
        x: torch.Tensor,
    ) -> tuple[torch.Tensor]:
        r"""Compute the squared L2 norm of the input tensor.

        Calculates :math:`\| W * (x - b) \|_2^2`, where :math:`W` is `weight` and :math`b` is `target`.
        The squared norm is computed along dimensions specified by `dim`.
        If `divide_by_n` is `True`, the result is averaged over these
        dimensions; otherwise, it's summed.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            The squared L2 norm. If `keepdim` is `True`, the dimensions `dim` are retained
            with size 1; otherwise, they are reduced.
        """
        return super().__call__(x)

    def forward(
        self,
        x: torch.Tensor,
    ) -> tuple[torch.Tensor]:
        """Apply forward of L2NormSquared.

        .. note::
            Prefer calling the instance of the L2NormSquared as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        value = (self.weight * (x - self.target)).abs().square()

        if self.divide_by_n:
            return (torch.mean(value, dim=self.dim, keepdim=self.keepdim),)
        else:
            return (torch.sum(value, dim=self.dim, keepdim=self.keepdim),)

    def prox(
        self,
        x: torch.Tensor,
        sigma: torch.Tensor | float = 1.0,
    ) -> tuple[torch.Tensor]:
        """Proximal Mapping of the squared L2 Norm.

        Apply the proximal mapping of the squared L2 norm.

        Parameters
        ----------
        x
            input tensor
        sigma
            scaling factor

        Returns
        -------
            Proximal mapping applied to the input tensor
        """
        throw_if_negative_or_complex(sigma)
        weight_square_2_sigma = self._divide_by_n(
            self.weight.conj() * self.weight * 2 * sigma,
            torch.broadcast_shapes(x.shape, self.target.shape, self.weight.shape),
        )
        x_out = (x + weight_square_2_sigma * self.target) / (1.0 + weight_square_2_sigma)

        return (x_out,)

    def prox_convex_conj(
        self,
        x: torch.Tensor,
        sigma: torch.Tensor | float = 1.0,
    ) -> tuple[torch.Tensor]:
        """Convex conjugate of squared L2 Norm.

        Apply the proximal mapping of the convex conjugate of the squared L2 norm.

        Parameters
        ----------
        x
            data tensor
        sigma
            scaling factor

        Returns
        -------
            Proximal of convex conjugate applied to the input tensor
        """
        throw_if_negative_or_complex(sigma)
        weight_square = self._divide_by_n(
            self.weight.conj() * self.weight, torch.broadcast_shapes(x.shape, self.target.shape, self.weight.shape)
        )

        x_out = (2 * weight_square * (x - sigma * self.target)) / (sigma + 2 * weight_square)
        return (x_out,)



"""Modified Look-Locker inversion recovery (MOLLI) signal model for T1 mapping."""

from collections.abc import Sequence

import torch

from mrpro.operators.SignalModel import SignalModel
from mrpro.utils.reshape import unsqueeze_right


class MOLLI(SignalModel[torch.Tensor, torch.Tensor, torch.Tensor]):
    r"""Signal model for Modified Look-Locker inversion recovery (MOLLI).

    This model describes
    :math:`M_z(t) = a(1 - (1 + c) e^{-t / T_1^*})` with :math:`T_1^* = T_1 / c`.

    This is a small modification from the original MOLLI signal model [MESS2004]_:
    :math:`M_z(t) = a - be^{(-t / T_1^*)}` with :math:`T_1^* = T1 / (b/a - 1)`.

    For a meaningful result chose :math:`c \in R_{>0}`, :math:`t \in R_{>0}`, and :math:`T_1 \in R_{>0}`

    .. [MESS2004] Messroghli DR, Radjenovic A, Kozerke S, Higgins DM, Sivananthan MU, Ridgway JP (2004) Modified
      look-locker inversion recovery (MOLLI) for high-resolution T 1 mapping of the heart. MRM, 52(1).
      https://doi.org/10.1002/mrm.20110
    """

    def __init__(self, ti: float | torch.Tensor | Sequence[float]):
        """Initialize MOLLI signal model for T1 mapping.

        Parameters
        ----------
        ti
            inversion times
            with shape (time, ...)
        """
        super().__init__()
        ti = torch.as_tensor(ti)
        self.ti = torch.nn.Parameter(ti, requires_grad=ti.requires_grad)

    def __call__(self, a: torch.Tensor, c: torch.Tensor, t1: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the Modified Look-Locker Inversion recovery (MOLLI) signal model.

        Calculates the signal based on the formula:
        :math:`S(TI) = a(1 - (1 + c) e^{-TI c / T1})`,
        where `TI` are the inversion times.

        Parameters
        ----------
        a
            Parameter 'a' in the MOLLI signal model.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        c
            Parameter 'c = b/a' in the MOLLI signal model.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        t1
            Longitudinal relaxation time T1.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.

        Returns
        -------
            Shape `(times ...)`, for example `(times, *other, coils, z, y, x)`, or `(times, samples)`
            where `times` is the number of inversion times.
        """
        return super().__call__(a, c, t1)

    def forward(self, a: torch.Tensor, c: torch.Tensor, t1: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of MOLLI.

        .. note::
            Prefer calling the instance of the MOLLI as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        ndim = max(a.ndim, c.ndim, t1.ndim)
        ti = unsqueeze_right(self.ti, ndim - self.ti.ndim + 1)  # leftmost is time
        signal = a * (1 - (1 + c) * torch.exp(-ti / t1 * c))
        return (signal,)



"""Sampling from von Mises-Fisher distribution."""

# based on: https://github.com/jasonlaska/spherecluster/blob/701b0b1909088a56e353b363b2672580d4fe9d93/spherecluster/util.py
# http://stats.stackexchange.com/questions/156729/sampling-from-von-mises-fisher-distribution-in-python
# https://www.mitsuba-renderer.org/~wenzel/files/vmf.pdf
# http://www.stat.pitt.edu/sungkyu/software/randvonMisesFisher3.pdf

from math import log, sqrt

import torch


def sample_vmf(mu: torch.Tensor, kappa: float, n_samples: int) -> torch.Tensor:
    """
    Generate samples from the von Mises-Fisher distribution.

    The von Mises-Fisher distribution is a circular normal distribution on the unit hypersphere.

    Parameters
    ----------
    mu
        Center of the distribution on the unit hypersphere. Shape: (..., dim)
    kappa
        Concentration parameter.
        For small kappa, the distribution is close to uniform.
        For large kappa, the distribution is close to a normal distribution with variance 1/kappa.
    n_samples
        Number of samples to generate.

    Returns
    -------
        Samples from the von Mises-Fisher distribution. Shape: (num_samples, ..., dim)
    """
    mu_ = mu.unsqueeze(0) if mu.dim() == 1 else mu
    total_samples = n_samples * mu_[..., 0].numel()
    mu_ = mu_.expand((n_samples, *mu_.shape))
    dim = mu_.shape[-1]

    b = (dim - 1) / (sqrt(4.0 * kappa**2 + (dim - 1) ** 2) + 2 * kappa)
    x = (1.0 - b) / (1.0 + b)
    c = kappa * x + (dim - 1) * log(1 - x**2)

    beta_dist = torch.distributions.Beta((dim - 1) / 2.0, (dim - 1) / 2.0)
    uniform_dist = torch.distributions.Uniform(0, 1)
    normal_dist = torch.distributions.Normal(0, 1)

    ws: list[torch.Tensor] = []

    while sum(len(w) for w in ws) < total_samples:
        # rejection sampling
        z = beta_dist.sample(torch.Size((total_samples,)))
        w = (1.0 - (1.0 + b) * z) / (1.0 - (1.0 - b) * z)
        u = uniform_dist.sample(torch.Size((total_samples,)))
        accepted = kappa * w + (dim - 1) * torch.log(1.0 - x * w) - c >= torch.log(u)
        ws.append(w[accepted])
    weights = torch.cat(ws)[:total_samples].reshape(mu_.shape[:-1])

    v = normal_dist.sample(mu_.shape)
    orthogonal_vectors = v - (mu_ * v).sum(-1, keepdim=True) * mu_ / mu_.norm(dim=-1, keepdim=True)
    orthonormal_vectors = orthogonal_vectors / orthogonal_vectors.norm(dim=-1, keepdim=True)
    samples = orthonormal_vectors * (1.0 - weights**2).sqrt().unsqueeze(-1) + weights.unsqueeze(-1) * mu_
    if mu.dim() == 1:
        samples = samples.squeeze(-2)
    return samples



"""Patch extraction (sliding window) operator."""

from collections.abc import Sequence

import torch

from mrpro.operators.LinearOperator import LinearOperator
from mrpro.utils.sliding_window import sliding_window


class PatchOp(LinearOperator):
    """Extract N-dimensional patches using a sliding window view.

    The adjoint assembles patches to an image.
    """

    def __init__(
        self,
        dim: Sequence[int] | int,
        patch_size: Sequence[int] | int,
        stride: Sequence[int] | int | None = None,
        dilation: Sequence[int] | int = 1,
        domain_size: int | Sequence[int] | None = None,
    ) -> None:
        """Initialize the PatchOp.

        Parameters
        ----------
        dim
            Dimension(s) to extract patches from.
        patch_size
            Size of patches (window_shape).
        stride
            Stride between patches. Set to `patch_size` if None.
        dilation
            Dilation factor of the patches
        domain_size
            Size of the domain in the dimnsions `dim`.
            If None, it is inferred from the input tensor on the first call.
            This is only used in the adjoint method.
        """
        super().__init__()
        self.dim = (dim,) if isinstance(dim, int) else dim

        if len(set(self.dim)) != len(self.dim):
            raise ValueError('Duplicate values in axis are not allowed')

        def check(param: int | Sequence[int], name: str) -> tuple[int, ...]:
            if isinstance(param, int):
                param = (param,) * len(self.dim)
            elif len(param) != len(self.dim):
                raise ValueError(f'Length mismatch: {name} must have length {len(self.dim)}')
            else:
                param = tuple(param)
            if any(val <= 0 for val in param):
                raise ValueError(f'{name} must be positive')
            return param

        self.patch_size = check(patch_size, 'patch_size')
        self.stride = check(stride, 'stride') if stride is not None else self.patch_size
        self.dilation = check(dilation, 'dilation')
        self.domain_size = check(domain_size, 'domain_size') if domain_size is not None else None

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Extract N-dimensional patches from an input tensor using a sliding window.

        Parameters
        ----------
        x
            Input tensor from which to extract patches.

        Returns
        -------
            A tensor containing the extracted patches. The first dimension
            represents the number of patches, followed by the original
            tensor dimensions (excluding those used for patching), and then
            the patch dimensions themselves.
            Shape: `(n_patches, ... , patch_size_dim1, patch_size_dim2, ...)`.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of PatchOp.

        .. note::
            Prefer calling the instance of the PatchOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        domain_size = tuple(x.shape[dim] for dim in self.dim)
        if self.domain_size is None:
            self.domain_size = domain_size
        elif tuple(self.domain_size) != domain_size:
            raise ValueError(
                f'Domain size {self.domain_size} does not match input shape in dimensions {self.dim}: {domain_size}'
            )
        patches = sliding_window(
            x=x,
            window_shape=self.patch_size,
            dim=self.dim,
            stride=self.stride,
            dilation=self.dilation,
        )
        patches = patches.flatten(start_dim=0, end_dim=len(self.dim) - 1)
        return (patches,)

    def adjoint(
        self,
        patches: torch.Tensor,
    ) -> tuple[torch.Tensor,]:
        """Assemble patches back into an image (adjoint operation).

        This method reconstructs an image by summing the provided patches
        at their respective locations, effectively reversing the patch
        extraction process. Overlapping areas are summed.

        Parameters
        ----------
        patches
            Tensor of patches to be assembled. Expected shape is
            `(n_patches, ..., patch_size_dim1, patch_size_dim2, ...)`.

        Returns
        -------
            The assembled image. Its shape will match the original image
            from which patches would have been extracted, with patch dimensions
            replaced by the original domain sizes along those dimensions.
        """
        if self.domain_size is None:
            raise ValueError('Domain size is not set. Please call forward first or set it at initialization.')

        output_shape_ = list(patches.shape[1:])
        for dim, size in zip(self.dim, self.domain_size, strict=True):
            output_shape_[dim] = size
        output_shape = torch.Size(output_shape_)
        indices = torch.arange(output_shape.numel(), device=patches.device).reshape(output_shape_)
        windowed_indices = sliding_window(
            x=indices,
            window_shape=self.patch_size,
            dim=self.dim,
            stride=self.stride,
            dilation=self.dilation,
        ).flatten(start_dim=0, end_dim=len(self.dim) - 1)
        if windowed_indices.shape[0] != patches.shape[0]:
            raise ValueError(
                f'Number of patches {patches.shape[0]} does not match the number of '
                f'expected patches {windowed_indices.shape[0]}'
            )

        assembled = patches.new_zeros(output_shape.numel())
        assembled.scatter_add_(dim=0, index=windowed_indices.flatten(), src=patches.flatten())
        assembled = assembled.reshape(output_shape)
        return (assembled,)



"""FastMRI dataset."""

import re
from collections import defaultdict
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
from warnings import warn
from xml.etree import ElementTree as ET

import h5py
import numpy as np
import torch
from einops import rearrange

from mrpro.data.AcqInfo import AcqInfo
from mrpro.data.enums import TrajectoryType
from mrpro.data.KData import KData
from mrpro.data.KHeader import KHeader
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.data.traj_calculators.KTrajectoryCartesian import KTrajectoryCartesian
from mrpro.utils.reshape import unsqueeze_left
from mrpro.utils.unit_conversion import deg_to_rad, magnetic_field_to_lamor_frequency, ms_to_s


class M4RawDataset(torch.utils.data.Dataset):
    """M4Raw KData Dataset.

    M4Raw [M4RAW]_ low-field (0.3T) brain dataset. The data can be downloaded from zenodo [M4RAW_DATA]_.
    The data consists of Cartsian GRE, T1w, T2w, and FLAIR images with multiple repetitions.
    It will be returned as `mrpro.data.KData` objects of either a single slice or a stack of
    slices with multiple repetitions.

    References
    ----------
    .. [M4RAW_DATA] Lyu, M., Mei, L., Huang, S., et al.(2023). M4Raw: A multi-contrast, multi-repetition,
       multi-channel MRI k-space dataset for low-field MRI research [Data set] https://doi.org/10.5281/zenodo.8056074
    .. [M4RAW] Lyu, M., Mei, L., Huang, S. et al. M4Raw: A multi-contrast, multi-repetition, multi-channel MRI k-space
       dataset for low-field MRI research. Sci Data 10, 264 (2023). https://doi.org/10.1038/s41597-023-02181-4
    """

    def __init__(self, path: PathLike | str | Sequence[PathLike | str], single_slice: bool = False):
        """Initialize the dataset.

        Parameters
        ----------
        path : PathLike
            Either a path to a directory containing the M4Raw data as .h5 files or a sequence of paths of
            individual files.
        single_slice : bool
            Whether to return single slices or stacks of slices.
        """
        pattern = re.compile(r'(\d+_(?:FLAIR|T1|T2|GRE))(\d+)')
        grouped: dict[str, list] = defaultdict(list)
        for fn in Path(path).rglob('*.h5') if isinstance(path, str | Path | PathLike) else [Path(p) for p in path]:
            m = pattern.fullmatch(fn.stem)
            if m is None:
                warn(f'Ignoring file {fn} because it does not match pattern.', stacklevel=2)
                continue
            grouped[m[1]].append(fn)
        self._filenames = list(grouped.values())
        self.single_slice = single_slice

    def __len__(self) -> int:
        """Get length (number of slices or stacks of slices) of the dataset."""
        if self.single_slice:
            return len(self._filenames) * 18
        else:
            return len(self._filenames)

    def __getitem__(self, idx: int) -> KData:
        """Get a single slice or stack of slices."""
        if not -len(self) <= idx < len(self):
            raise IndexError(f'Index {idx} is out of bounds for the dataset of size {len(self)}')
        if idx < 0:
            idx += len(self)

        if self.single_slice:
            slice_idx: int | slice
            file_idx, slice_idx = divmod(idx, 18)
        else:
            file_idx = idx
            slice_idx = slice(None)

        with h5py.File(self._filenames[file_idx][0], 'r') as file:
            xml_root = ET.fromstring(file['ismrmrd_header'][()].decode('utf-8'))  # noqa: S314

        info = AcqInfo()
        info.idx.k1 = torch.arange(30, 225)[None, None, None, :, None]
        # The dataset includes a header with an ISMRMRD namespace. But the XML is not a valid ISMRMRD header.
        # So we need to parse it manually.
        recon_matrix = encoding_matrix = SpatialDimension(1, 256, 256)  # Fix for all files
        recon_fov = encoding_fov = SpatialDimension(0.005, 0.24, 0.24)  # Fix for all files

        def get(name: str) -> list[str]:
            return [
                e.text for e in xml_root.findall(name, {'ns0': 'http://www.ismrm.org/ISMRMRD'}) if e.text is not None
            ]

        te = [ms_to_s(float(e)) for e in get('ns0:sequenceParameters/ns0:TE')]
        ti = [ms_to_s(float(e)) for e in get('ns0:sequenceParameters/ns0:TI')]
        fa = [deg_to_rad(float(e)) for e in get('ns0:sequenceParameters/ns0:flipAngle_deg')]
        tr = [ms_to_s(float(e)) for e in get('ns0:sequenceParameters/ns0:TR')]
        echo_spacing = [
            ms_to_s(float(e.replace('ms', ''))) for e in get('ns0:sequenceParameters/ns0:echo_spacing') if e != 'N/A'
        ]
        echo_train_length = int(get('ns0:encoding/ns0:echoTrainLength')[0])
        sequence_type = get('ns0:sequenceParameters/ns0:sequence_type')[0]
        model = get('ns0:acquisitionSystemInformation/ns0:systemModel')[0]
        vendor = get('ns0:acquisitionSystemInformation/ns0:systemVendor')[0]
        protocol_name = get('ns0:measurementInformation/ns0:protocolName')[0]
        measurement_id = get('ns0:measurementInformation/ns0:measurementID')[0]

        header = KHeader(
            recon_matrix=recon_matrix,
            encoding_matrix=encoding_matrix,
            recon_fov=recon_fov,
            encoding_fov=encoding_fov,
            te=te,
            ti=ti,
            fa=fa,
            tr=tr,
            echo_spacing=echo_spacing,
            echo_train_length=echo_train_length,
            sequence_type=sequence_type,
            model=model,
            vendor=vendor,
            protocol_name=protocol_name,
            measurement_id=measurement_id,
            trajectory_type=TrajectoryType.CARTESIAN,
            acq_info=info,
            lamor_frequency_proton=int(magnetic_field_to_lamor_frequency(0.3)),
        )

        reps = []
        for filename in self._filenames[file_idx]:
            with h5py.File(filename, 'r') as file:
                reps.append(torch.as_tensor(np.array(file['kspace'][slice_idx, ..., 30:225])))
        data = rearrange(torch.stack(reps), 'reps ... coils k0 k1 -> ... reps coils 1 k1 k0')
        info = AcqInfo()
        info.idx.k1 = unsqueeze_left(torch.arange(30, 225)[:, None], data.ndim - 2)
        info.idx.repetition = unsqueeze_left(torch.arange(data.shape[-5])[:, None, None, None, None], data.ndim - 5)

        traj = KTrajectoryCartesian()(
            n_k0=256,
            k0_center=128,
            k1_idx=info.idx.k1,
            k1_center=128,
            k2_idx=torch.tensor(0),
            k2_center=0,
        )
        return KData(data=data, header=header, traj=traj)



"""qMRI signal models."""

from mrpro.operators.models.SaturationRecovery import SaturationRecovery
from mrpro.operators.models.InversionRecovery import InversionRecovery
from mrpro.operators.models.SpoiledGRE import SpoiledGRE
from mrpro.operators.models.MOLLI import MOLLI
from mrpro.operators.models.WASABI import WASABI
from mrpro.operators.models.WASABITI import WASABITI
from mrpro.operators.models.MonoExponentialDecay import MonoExponentialDecay
from mrpro.operators.models.cMRF import CardiacFingerprinting
from mrpro.operators.models.TransientSteadyStateWithPreparation import TransientSteadyStateWithPreparation
from mrpro.operators.models import EPG

__all__ = [
    "CardiacFingerprinting",
    "EPG",
    "InversionRecovery",
    "MOLLI",
    "MonoExponentialDecay",
    "SaturationRecovery",
    "SpoiledGRE",
    "TransientSteadyStateWithPreparation",
    "WASABI",
    "WASABITI"
]



"""Mono-exponential decay signal model."""

from collections.abc import Sequence

import torch

from mrpro.operators.SignalModel import SignalModel
from mrpro.utils.reshape import unsqueeze_right


class MonoExponentialDecay(SignalModel[torch.Tensor, torch.Tensor]):
    """Signal model for mono-exponential decay."""

    def __init__(self, decay_time: float | torch.Tensor | Sequence[float]):
        """Initialize mono-exponential signal model.

        Can, for example, be used to model T2.

        Parameters
        ----------
        decay_time
            time points when model is evaluated
            with shape (time, ...)
        """
        super().__init__()
        decay_time = torch.as_tensor(decay_time)
        self.decay_time = torch.nn.Parameter(decay_time, requires_grad=decay_time.requires_grad)

    def __call__(self, m0: torch.Tensor, decay_constant: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the mono-exponential decay signal model.

        Calculates the signal based on the formula:
        :math:`S(t) = M_0 e^{-t / T}`,
        where `t` are the decay times and `T` is the decay constant.

        Parameters
        ----------
        m0
            Equilibrium signal / proton density.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        decay_constant
            Exponential decay constant (e.g., T2, T2*, T1rho).
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.

        Returns
        -------
            Signal calculated for each decay time.
            Shape `(times ...)`, for example `(times, *other, coils, z, y, x)`, or `(times, samples)`
            where `times` is the number of decay times.
        """
        return super().__call__(m0, decay_constant)

    def forward(self, m0: torch.Tensor, decay_constant: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of MonoExponentialDecay.

        .. note::
            Prefer calling the instance of the MonoExponentialDecay as ``operator(x)`` over directly calling this
            method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        ndim = max(m0.ndim, decay_constant.ndim)
        decay_time = unsqueeze_right(self.decay_time, ndim - self.decay_time.ndim + 1)  # leftmost is time
        signal = m0 * torch.exp(-(decay_time / decay_constant))
        return (signal,)



"""Operator enforcing constraints by variable transformations."""

from collections.abc import Sequence

import torch

from mrpro.operators.EndomorphOperator import EndomorphOperator, endomorph


def sigmoid(x: torch.Tensor, beta: float = 1.0) -> torch.Tensor:
    """Beta scaled sigmoid function."""
    return torch.nn.functional.sigmoid(beta * x)


def sigmoid_inverse(x: torch.Tensor, beta: float = 1.0) -> torch.Tensor:
    """Inverse of `sigmoid`."""
    return torch.logit(x) / beta


def softplus(x: torch.Tensor, beta: float = 1.0) -> torch.Tensor:
    """Beta scaled softplus function."""
    return -(1 / beta) * torch.nn.functional.logsigmoid(-beta * x)


def softplus_inverse(x: torch.Tensor, beta: float = 1.0) -> torch.Tensor:
    """Inverse of `softplus`."""
    return x + torch.log(-torch.expm1(-beta * x)) / beta


class ConstraintsOp(EndomorphOperator):
    """Transformation to map real-valued tensors to certain ranges."""

    def __init__(
        self,
        bounds: Sequence[tuple[float | None, float | None]],
        beta_sigmoid: float = 1.0,
        beta_softplus: float = 1.0,
    ) -> None:
        """Initialize a constraint operator.

        The operator maps real-valued tensors to certain ranges. The transformation is applied element-wise.
        The transformation is defined by the bounds. The bounds are applied in the order of the input tensors.
        If there are more input tensors than bounds, the remaining tensors are passed through without transformation.

        If an input tensor is bounded from below AND above, a sigmoid transformation is applied.
        If an input tensor is bounded from below OR above, a softplus transformation is applied.

        If an input is complex valued, the bounds are to the real and imaginary parts separately,
        i.e., for bounds :math:`(a, b)`, the complex number is constrained to a rectangle in the complex plane
        with corners :math:`(a+ai, a+bi, b+ai, b+bi)`.

        Parameters
        ----------
        bounds
            Sequence of `(lower_bound, upper_bound)` values. If a bound is `None`, the value is not constrained.
            If a lower bound is `-inf`, the value is not constrained from below. If an upper bound is `+inf`,
            the value is not constrained from above.
            If the bounds are set to `(None, None)` or `(-inf, +inf)`, the value is not constrained at all.
        beta_sigmoid
            beta parameter for the sigmoid transformation (used if an input has two bounds).
            A higher value leads to a steeper sigmoid.
        beta_softplus
            parameter for the softplus transformation (used if an input is either bounded from below or above).
            A higher value leads to a steeper softplus.

        Raises
        ------
        ValueError
            If the lower bound is greater than the upper bound.
        ValueError
            If the a bound is nan.
        ValueError
            If the parameter beta_sigmoid and beta_softplus are not greater than zero.
        """
        super().__init__()

        if beta_sigmoid <= 0:
            raise ValueError(f'parameter beta_sigmoid must be greater than zero; given {beta_sigmoid}')
        if beta_softplus <= 0:
            raise ValueError(f'parameter beta_softplus must be greater than zero; given {beta_softplus}')

        self.beta_sigmoid = beta_sigmoid
        self.beta_softplus = beta_softplus

        self.lower_bounds = tuple(torch.as_tensor(-torch.inf if lb is None else lb) for (lb, ub) in bounds)
        self.upper_bounds = tuple(torch.as_tensor(torch.inf if ub is None else ub) for (lb, ub) in bounds)

        for lb, ub in zip(self.lower_bounds, self.upper_bounds, strict=True):
            if lb.isnan():
                raise ValueError('nan is invalid as lower bound.')
            if ub.isnan():
                raise ValueError('nan is invalid as upper bound.')
            if lb >= ub:
                raise ValueError(
                    'bounds should be ( (a1,b1), (a2,b2), ...) with ai < bi if neither ai or bi is None;'
                    f'\nbound tuple {lb, ub} is invalid as the lower bound is higher than the upper bound',
                )

    def _apply_forward(self, item: torch.Tensor, lb: torch.Tensor, ub: torch.Tensor) -> torch.Tensor:
        """Apply the forward transformation to the input tensor."""
        if item.dtype.is_complex:
            real = self._apply_forward(item.real, lb, ub)
            imag = self._apply_forward(item.imag, lb, ub)
            return torch.complex(real, imag)

        if not lb.isneginf() and not ub.isposinf():
            # bounds are (lb,ub)
            return lb + (ub - lb) * sigmoid(item, beta=self.beta_sigmoid)

        if not lb.isneginf():
            # bounds are (lb,inf)
            return lb + softplus(item, beta=self.beta_softplus)

        if not ub.isposinf():
            # bounds are (-inf,ub)
            return ub - softplus(-item, beta=self.beta_softplus)

        return item  # unconstrained case

    def _apply_inverse(self, item: torch.Tensor, lb: torch.Tensor, ub: torch.Tensor) -> torch.Tensor:
        if item.dtype.is_complex:
            real = self._apply_inverse(item.real, lb, ub)
            imag = self._apply_inverse(item.imag, lb, ub)
            return torch.complex(real, imag)

        if not lb.isneginf() and not ub.isposinf():
            # bounds are (lb,ub)
            return sigmoid_inverse((item - lb) / (ub - lb), beta=self.beta_sigmoid)

        if not lb.isneginf():
            # bounds are (lb,inf)
            return softplus_inverse(item - lb, beta=self.beta_softplus)

        if not ub.isposinf():
            # bounds are (-inf,ub)
            return -softplus_inverse(-(item - ub), beta=self.beta_softplus)

        return item  # unconstrained case

    @endomorph
    def __call__(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Transform tensors to chosen range.

        Applies element-wise transformations to map input tensors to specified bounds.
        - If bounded below and above: uses a sigmoid transformation.
        - If bounded below or above: uses a softplus transformation.
        - If complex: applies transformation to real and imaginary parts separately.
        - If more input tensors than bounds: remaining tensors pass through unchanged.

        Parameters
        ----------
        *x
            One or more input tensors to be transformed.

        Returns
        -------
            Transformed tensors, with values mapped to the ranges defined by the bounds.
        """
        return super().__call__(*x)

    @endomorph
    def forward(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply forward of ConstraintsOp.

        .. note::
            Prefer calling the instance of the ConstraintsOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        x_constrained = [
            self._apply_forward(item, lb, ub)
            for item, lb, ub in zip(x, self.lower_bounds, self.upper_bounds, strict=False)
        ]
        # if there are more inputs than bounds, pass on the remaining inputs without transformation
        x_constrained.extend(x[len(x_constrained) :])
        return tuple(x_constrained)

    @endomorph
    def invert(self, *x_constrained: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Reverses the variable transformation.

        Parameters
        ----------
        x_constrained
            transformed tensors with values in the range defined by the bounds

        Returns
        -------
            tensors in the domain with no bounds
        """
        x = [
            self._apply_inverse(item, lb, ub)
            for item, lb, ub in zip(x_constrained, self.lower_bounds, self.upper_bounds, strict=False)
        ]
        # if there are more inputs than bounds, pass on the remaining inputs without transformation
        x.extend(x_constrained[len(x) :])

        return tuple(x)

    @property
    def inverse(self) -> 'InverseConstraintOp':
        """Return the inverse of the constraint operator."""
        return InverseConstraintOp(self)


class InverseConstraintOp(EndomorphOperator):
    """Inverse of a constraint operator."""

    def __init__(self, constraints_op: ConstraintsOp) -> None:
        """Initialize the inverse constraint operator.

        Parameters
        ----------
        constraints_op
            The constraint operator to invert.
        """
        super().__init__()
        self.constraints_op = constraints_op

    @endomorph
    def __call__(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply the inverse of the constraint operator.

        This reverses the transformation applied by the corresponding `ConstraintsOp`,
        mapping values from their constrained ranges back to the unbounded domain.

        Parameters
        ----------
        *x
            One or more input tensors, assumed to be in the constrained ranges.

        Returns
        -------
            Tensors with the inverse transformation applied, mapped back to the unbounded domain.
        """
        return super().__call__(*x)

    @endomorph
    def forward(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply forward of InverseConstraintOp.

        .. note::
            Prefer calling the instance of the InverseConstraintOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return self.constraints_op.invert(*x)

    @endomorph
    def invert(self, *x_constrained: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply constraint operator."""
        return self.constraints_op.forward(*x_constrained)

    @property
    def inverse(self) -> 'ConstraintsOp':
        """Return the constraint operator."""
        return self.constraints_op



"""Direct Reconstruction by Adjoint Fourier Transform."""

from collections.abc import Callable

from mrpro.algorithms.reconstruction.Reconstruction import Reconstruction
from mrpro.data.CsmData import CsmData
from mrpro.data.DcfData import DcfData
from mrpro.data.IData import IData
from mrpro.data.KData import KData
from mrpro.data.KNoise import KNoise
from mrpro.operators.FourierOp import FourierOp
from mrpro.operators.LinearOperator import LinearOperator


class DirectReconstruction(Reconstruction):
    """Direct Reconstruction by Adjoint Fourier Transform."""

    def __init__(
        self,
        kdata: KData | None = None,
        fourier_op: LinearOperator | None = None,
        csm: Callable[[IData], CsmData] | CsmData | None = CsmData.from_idata_walsh,
        noise: KNoise | None = None,
        dcf: DcfData | None = None,
    ):
        """Initialize DirectReconstruction.

        A direct reconstruction uses the adjoint of the acquisition operator and a
        density compensation to obtain the complex valued images from k-space data.

        If csm is not set to `None`, a single coil combined image will reconstructed.
        The method for estimating sensitivity maps can be adjusted using the `csm` argument.

        Parameters
        ----------
        kdata
            If `kdata` is provided and `fourier_op` or `dcf` are `None`, then `fourier_op` and `dcf` are estimated
            based on `kdata`. Otherwise `fourier_op` and `dcf` are used as provided.
        fourier_op
            Instance of the `~mrpro.operators.FourierOp` used for reconstruction.
            If `None`, set up based on `kdata`.
        csm
            Sensitivity maps for coil combination. If `None`, no coil combination is carried out, i.e. images for each
            coil are returned. If a `Callable` is provided, coil images are reconstructed using the adjoint of the
            `~mrpro.operators.FourierOp` (including density compensation) and then sensitivity maps are calculated
            using the callable. For this, `kdata` needs also to be provided.
            For examples have a look at the `~mrpro.data.CsmData` class e.g. `~mrpro.data.CsmData.from_idata_walsh`
            or `~mrpro.data.CsmData.from_idata_inati`.
        noise
            Noise used for prewhitening. If `None`, no prewhitening is performed
        dcf
            K-space sampling density compensation. If `None`, set up based on `kdata`.

        Raises
        ------
        `ValueError`
            If the `kdata` and `fourier_op` are `None` or if `csm` is a `Callable` but `kdata` is None.
        """
        super().__init__()
        if fourier_op is None:
            if kdata is None:
                raise ValueError('Either kdata or fourier_op needs to be defined.')
            self.fourier_op = FourierOp.from_kdata(kdata)
        else:
            self.fourier_op = fourier_op

        if kdata is not None and dcf is None:
            self.dcf = DcfData.from_traj_voronoi(kdata.traj)
        else:
            self.dcf = dcf

        self.noise = noise

        if csm is None or isinstance(csm, CsmData):
            self.csm = csm
        else:
            if kdata is None:
                raise ValueError('kdata needs to be defined to calculate the sensitivity maps.')
            self.recalculate_csm(kdata, csm)

    def forward(self, kdata: KData) -> IData:
        """Apply the reconstruction.

        Parameters
        ----------
        kdata
            k-space data to reconstruct.

        Returns
        -------
            the reconstruced image.
        """
        return self.direct_reconstruction(kdata)



"""Inversion recovery signal model for T1 mapping."""

from collections.abc import Sequence

import torch

from mrpro.operators.SignalModel import SignalModel
from mrpro.utils.reshape import unsqueeze_right


class InversionRecovery(SignalModel[torch.Tensor, torch.Tensor]):
    """Inversion recovery signal model."""

    def __init__(self, ti: float | torch.Tensor | Sequence[float]):
        """Initialize inversion recovery signal model for T1 mapping.

        Parameters
        ----------
        ti
            inversion times
            with shape (time, ...)
        """
        super().__init__()
        ti = torch.as_tensor(ti)
        self.ti = torch.nn.Parameter(ti, requires_grad=ti.requires_grad)

    def __call__(self, m0: torch.Tensor, t1: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the inversion recovery signal model.

        Calculates the signal based on the formula:
        :math:`S(TI) = M_0 (1 - 2 * e^{-TI / T_1})`,
        where `TI` are the inversion times.

        Parameters
        ----------
        m0
            Equilibrium signal / proton density.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        t1
            Longitudinal relaxation time T1.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.

        Returns
        -------
            Signal calculated for each inversion time.
            Shape `(times ...)`, for example `(times, *other, coils, z, y, x)`, or `(times, samples)`
            where `times` is the number of inversion times.
        """
        return super().__call__(m0, t1)

    def forward(self, m0: torch.Tensor, t1: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of InversionRecovery.

        .. note::
            Prefer calling the instance of the InversionRecovery as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        ndim = max(m0.ndim, t1.ndim)
        ti = unsqueeze_right(self.ti, ndim - self.ti.ndim + 1)  # leftmost is time
        signal = m0 * (1 - 2 * torch.exp(-(ti / t1)))
        return (signal,)
